from django.db import models

from canvas_sdk.v1.data.base import IdentifiableModel, TimestampedModel


class TransmissionChannel(models.TextChoices):
    """Transmission channel."""

    MANUAL = "manual", "Manual"
    TEXT_MESSAGE = "sms", "Text Message"
    EMAIL = "email", "Email"
    NOOP = "noop", "No-op"


class Message(TimestampedModel, IdentifiableModel):
    """Message."""

    class Meta:
        db_table = "canvas_sdk_data_api_message_001"

    content = models.TextField()
    sender = models.ForeignKey(
        "v1.CanvasUser", on_delete=models.DO_NOTHING, related_name="sent_messages", null=True
    )
    recipient = models.ForeignKey(
        "v1.CanvasUser", on_delete=models.DO_NOTHING, related_name="received_messages", null=True
    )
    note = models.ForeignKey(
        "v1.Note", on_delete=models.DO_NOTHING, related_name="message", null=True
    )
    read = models.DateTimeField(null=True, blank=True)


class MessageAttachment(IdentifiableModel):
    """Message attachment."""

    class Meta:
        db_table = "canvas_sdk_data_api_messageattachment_001"

    file = models.TextField()
    content_type = models.CharField(max_length=255)
    message = models.ForeignKey(
        "v1.Message", on_delete=models.DO_NOTHING, related_name="message", null=True
    )


class MessageTransmission(TimestampedModel, IdentifiableModel):
    """Message Transmission."""

    class Meta:
        db_table = "canvas_sdk_data_api_messagetransmission_001"

    message = models.ForeignKey(
        "v1.Message", on_delete=models.DO_NOTHING, related_name="transmissions", null=True
    )
    delivered = models.BooleanField()
    failed = models.BooleanField()

    contact_point_system = models.CharField(choices=TransmissionChannel.choices, max_length=20)
    contact_point_value = models.CharField(max_length=255)

    comment = models.TextField()
    delivered_by = models.ForeignKey(
        "v1.Staff",
        on_delete=models.DO_NOTHING,
        related_name="transmissions_delivered",
        null=True,
    )


__exports__ = ("TransmissionChannel", "Message", "MessageAttachment", "MessageTransmission")



# To run the tests, use the command `pytest` in the terminal or uv run pytest.
# Each test is wrapped inside a transaction that is rolled back at the end of the test.
# If you want to modify which files are used for testing, check the [tool.pytest.ini_options] section in pyproject.toml.


from canvas_sdk.test_utils.factories import PatientFactory
from canvas_sdk.v1.data.discount import Discount


# You can use a factory to create a patient instance for testing purposes.
def test_factory() -> None:
    """Test that a patient can be created using the PatientFactory."""
    patient = PatientFactory.create()
    assert patient.id is not None


# if a factory is not available, you can create an instance manually with the data model directly.
def test_model() -> None:
    """Test that a Discount instance can be created."""
    Discount.objects.create(name="10%", adjustment_group="30", adjustment_code="CO", discount=0.10)
    assert Discount.objects.first().pk is not None



import json

from canvas_generated.messages.effects_pb2 import Effect
from canvas_sdk.base import TrackableFieldsModel


class CreatePatientExternalIdentifier(TrackableFieldsModel):
    """Effect to create a Patient External Identifier record."""

    class Meta:
        effect_type = "PATIENT_EXTERNAL_IDENTIFIER"

    value: str
    system: str | None = None
    patient_id: str | None = None

    @property
    def values(self) -> dict[str, str | None]:
        """Return the values of the external identifier."""
        return {
            "value": self.value,
            "system": self.system,
            "patient_id": self.patient_id,
        }

    def create(self) -> Effect:
        """Create a new Patient External Identifier."""
        self._validate_before_effect("create")

        return Effect(
            type=f"CREATE_{self.Meta.effect_type}",
            payload=json.dumps(
                {
                    "data": self.values,
                }
            ),
        )


__exports__ = ("CreatePatientExternalIdentifier",)



from django.db import models

from canvas_sdk.v1.data.base import AuditedModel, IdentifiableModel


class StopMedicationEvent(AuditedModel, IdentifiableModel):
    """StopMedicationEvent."""

    class Meta:
        db_table = "canvas_sdk_data_api_stopmedicationevent_001"

    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="stopped_medications"
    )
    note = models.ForeignKey(
        "v1.Note", on_delete=models.DO_NOTHING, related_name="stopped_medications"
    )
    medication = models.ForeignKey(
        "v1.Medication",
        on_delete=models.DO_NOTHING,
        related_name="stopmedicationevent_set",
        null=True,
    )
    rationale = models.CharField(max_length=1024, default="")


__exports__ = ("StopMedicationEvent",)



from typing import Any

import factory

from canvas_sdk.v1.data import Note, NoteStateChangeEvent
from canvas_sdk.v1.data.note import NoteStates


class NoteFactory(factory.django.DjangoModelFactory[Note]):
    """Factory for creating Note."""

    class Meta:
        model = Note

    location = factory.SubFactory("canvas_sdk.test_utils.factories.PracticeLocationFactory")
    originator = factory.SelfAttribute("provider.user")
    provider = factory.SubFactory("canvas_sdk.test_utils.factories.StaffFactory")
    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")

    @factory.post_generation
    def create_initial_state_change_event(self, create: Any, extracted: Any, **kwargs: Any) -> None:
        """Create an initial NoteStateChangeEvent when a Note is created."""
        if not create:
            return

        NoteStateChangeEventFactory.create(note=self, state=NoteStates.NEW)


class NoteStateChangeEventFactory(factory.django.DjangoModelFactory[NoteStateChangeEvent]):
    """Factory for creating NoteStateChangeEvent."""

    class Meta:
        model = NoteStateChangeEvent

    note = factory.SubFactory(NoteFactory)
    state = NoteStates.NEW
    originator = factory.SelfAttribute("note.originator")



from datetime import date

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class MedicalHistoryCommand(BaseCommand):
    """A class for managing a Medical History command within a specific note."""

    class Meta:
        key = "medicalHistory"

    past_medical_history: str | None = None
    approximate_start_date: date | None = None
    approximate_end_date: date | None = None
    show_on_condition_list: bool = True
    comments: str | None = Field(max_length=1000, default=None)


__exports__ = ("MedicalHistoryCommand",)



from django.db import models

from canvas_sdk.v1.data.base import AuditedModel, IdentifiableModel


class GoalLifecycleStatus(models.TextChoices):
    """GoalLifecycleStatus choices."""

    PROPOSED = "proposed", "Proposed"
    PLANNED = "planned", "Planned"
    ACCEPTED = "accepted", "Accepted"
    ACTIVE = "active", "Active"
    ON_HOLD = "on-hold", "On Hold"
    COMPLETED = "completed", "Completed"
    CANCELLED = "cancelled", "Cancelled"
    REJECTED = "rejected", "Rejected"


class GoalAchievementStatus(models.TextChoices):
    """GoalAchievementStatus choices."""

    IN_PROGRESS = "in-progress", "In Progress"
    IMPROVING = "improving", "Improving"
    WORSENING = "worsening", "Worsening"
    NO_CHANGE = "no-change", "No Change"
    ACHIEVED = "achieved", "Achieved"
    SUSTAINING = "sustaining", "Sustaining"
    NOT_ACHIEVED = "not-achieved", "Not Achieved"
    NO_PROGRESS = "no-progress", "No Progress"
    NOT_ATTAINABLE = "not-attainable", "Not Attainable"


class GoalPriority(models.TextChoices):
    """GoalPriority choices."""

    HIGH = "high-priority", "High Priority"
    MEDIUM = "medium-priority", "Medium Priority"
    LOW = "low-priority", "Low Priority"


class Goal(AuditedModel, IdentifiableModel):
    """Goal."""

    class Meta:
        db_table = "canvas_sdk_data_api_goal_001"

    patient = models.ForeignKey("v1.Patient", on_delete=models.DO_NOTHING, related_name="goals")
    note = models.ForeignKey("v1.Note", on_delete=models.DO_NOTHING, related_name="goals")
    lifecycle_status = models.CharField(max_length=20, choices=GoalLifecycleStatus.choices)
    achievement_status = models.CharField(max_length=20, choices=GoalAchievementStatus.choices)
    priority = models.CharField(max_length=20, choices=GoalPriority.choices)
    due_date = models.DateField()
    progress = models.TextField()
    goal_statement = models.TextField()
    start_date = models.DateField()


__exports__ = ("Goal", "GoalLifecycleStatus", "GoalAchievementStatus", "GoalPriority")



import json
import random
from pathlib import Path
from typing import Annotated

import grpc
import typer

from canvas_generated.messages.events_pb2 import Event as PluginRunnerEvent
from canvas_generated.services.plugin_runner_pb2_grpc import PluginRunnerStub


def emit(
    event_fixture: str,
    plugin_runner_port: Annotated[
        str, typer.Option(help="Port of your locally running plugin runner")
    ] = "50051",
) -> None:
    """
    Send an event fixture to your locally running plugin-runner process, and print any resultant effects.

    Valid fixture files are newline-delimited JSON, with each containing the keys `EventType`, `target`, and `context`. Some fixture files are included in the canvas-plugins repo.
    """
    # If an event fixture file exists at the specified path, use it.
    # Otherwise, see if it represents an event that we have a Canvas-provided
    # fixture for and use that.
    event_fixture_path = Path(event_fixture)

    if not event_fixture_path.exists():
        candidate_built_in_fixture_path = (
            Path(__file__).resolve().parent / "event_fixtures" / f"{event_fixture}.ndjson"
        )
        if candidate_built_in_fixture_path.exists():
            event_fixture_path = candidate_built_in_fixture_path
        else:
            print(f"ERROR: No file found at location {event_fixture}.")
            print(f"ERROR: No built-in fixture file found named {event_fixture}.ndjson.")
            return

    # Grab a random event from the fixture file ndjson
    lines = event_fixture_path.read_text().splitlines()
    myline = random.choice(lines)
    event_data = json.loads(myline)
    event = PluginRunnerEvent(
        type=event_data["EventType"],
        target=event_data["target"],
        context=event_data["context"],
    )
    with grpc.insecure_channel(f"localhost:{plugin_runner_port}") as channel:
        stub = PluginRunnerStub(channel)
        responses = stub.HandleEvent(event)

        at_least_one_effect = False
        try:
            for response in responses:
                for effect in response.effects:
                    at_least_one_effect = True
                    print(effect)

            if not at_least_one_effect:
                print("SUCCESS: No effects returned.")
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                print(
                    f"ERROR: Couldn't make a connection to a plugin runner process at localhost:{plugin_runner_port}. Is it running?"
                )



from canvas_generated.messages import events_pb2 as _events_pb2
from canvas_generated.messages import plugins_pb2 as _plugins_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor



import json
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Any

from pydantic import Field, conlist
from pydantic_core import InitErrorDetails

from canvas_sdk.commands.base import _BaseCommand, _SendableCommandMixin
from canvas_sdk.commands.constants import ClinicalQuantity
from canvas_sdk.effects import Effect
from canvas_sdk.effects.compound_medications.compound_medication import (
    CompoundMedication as CompoundMedicationEffect,
)
from canvas_sdk.v1.data.compound_medication import CompoundMedication as CompoundMedicationModel


@dataclass
class CompoundMedicationData:
    """Data for creating a compound medication inline within a prescription."""

    formulation: str
    potency_unit_code: str
    controlled_substance: str
    controlled_substance_ndc: str = ""
    active: bool = True

    def to_dict(self) -> dict[str, Any]:
        """Convert dataclass to dictionary, excluding None values."""
        return {
            "formulation": self.formulation,
            "potency_unit_code": self.potency_unit_code,
            "controlled_substance": self.controlled_substance,
            "controlled_substance_ndc": self.controlled_substance_ndc,
            "active": self.active,
        }


class PrescribeCommand(_SendableCommandMixin, _BaseCommand):
    """A class for managing a Prescribe command within a specific note."""

    class Meta:
        key = "prescribe"

    class Substitutions(Enum):
        ALLOWED = "allowed"
        NOT_ALLOWED = "not_allowed"

    fdb_code: str | None = Field(default=None, json_schema_extra={"commands_api_name": "prescribe"})
    icd10_codes: conlist(str, max_length=2) = Field(  # type: ignore[valid-type]
        default=[], json_schema_extra={"commands_api_name": "indications"}
    )
    sig: str = ""
    days_supply: int | None = None
    quantity_to_dispense: Decimal | float | int | None = None
    type_to_dispense: ClinicalQuantity | None = None
    refills: int | None = None
    substitutions: Substitutions | None = None
    pharmacy: str | None = None
    prescriber_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "prescriber"}
    )
    supervising_provider_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "supervising_provider"}
    )
    note_to_pharmacist: str | None = None

    # Compound medication fields
    compound_medication_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "compound_medication_id"}
    )
    compound_medication_data: CompoundMedicationData | None = Field(
        default=None, json_schema_extra={"commands_api_name": "compound_medication_data"}
    )

    def _has_fdb_code(self) -> bool:
        """Check if fdb_code is provided and non-empty."""
        return self.fdb_code is not None and self.fdb_code.strip() != ""

    def _get_error_details(self, method: str) -> list[InitErrorDetails]:
        """Add compound medication validation to the base validation."""
        errors = super()._get_error_details(method)

        # Validate that exactly one medication type is provided
        has_fdb_code = self._has_fdb_code()
        has_compound_medication_id = (
            self.compound_medication_id is not None and self.compound_medication_id.strip() != ""
        )
        has_compound_medication_data = self.compound_medication_data is not None

        medication_types_provided = sum(
            [has_fdb_code, has_compound_medication_id, has_compound_medication_data]
        )

        if medication_types_provided > 1:
            errors.append(
                self._create_error_detail(
                    "value",
                    "Cannot specify multiple medication types. Choose one of: 'fdb_code', 'compound_medication_id', or 'compound_medication_data'.",
                    None,
                )
            )

        # Validate compound medication ID if provided
        if (
            self.compound_medication_id
            and self.compound_medication_id.strip() != ""
            and
            # Check if compound medication exists
            not CompoundMedicationModel.objects.filter(id=self.compound_medication_id).exists()
        ):
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Compound medication with ID {self.compound_medication_id} does not exist.",
                    self.compound_medication_id,
                )
            )

        # Validate compound medication data if provided
        if self.compound_medication_data is not None:
            compound_med_errors = CompoundMedicationEffect.validate_compound_medication_fields(
                formulation=self.compound_medication_data.formulation,
                potency_unit_code=self.compound_medication_data.potency_unit_code,
                controlled_substance=self.compound_medication_data.controlled_substance,
                controlled_substance_ndc=self.compound_medication_data.controlled_substance_ndc,
            )
            errors.extend(compound_med_errors)

        return errors

    def _process_compound_medication_fields(self, data: dict[str, Any]) -> dict[str, Any]:
        """
        Process compound medication fields in prescription data.

        Args:
            data: Dictionary containing prescription data

        Returns:
            Processed dictionary with compound medication transformations applied
        """
        # Process nested compound medication values if present
        if "compound_medication_values" in data:
            compound_med_data = data["compound_medication_values"]
            processed_compound_data = CompoundMedicationEffect.process_compound_medication_data(
                compound_med_data
            )
            data["compound_medication_values"] = processed_compound_data

        return data

    @property
    def values(self) -> dict:
        """The Prescribe command's field values."""
        values = super().values

        if self.is_dirty("quantity_to_dispense"):
            values["quantity_to_dispense"] = (
                str(Decimal(self.quantity_to_dispense)) if self.quantity_to_dispense else None
            )

        if self.is_dirty("compound_medication_id") and self.compound_medication_id:
            values["compound_medication_values"] = {"id": values.pop("compound_medication_id")}

        # Handle compound medication data
        elif (
            "compound_medication_data" in values and values["compound_medication_data"] is not None
        ):
            # Convert CompoundMedicationData to the expected nested structure
            compound_data = values.pop("compound_medication_data")
            if isinstance(compound_data, CompoundMedicationData):
                values["compound_medication_values"] = compound_data.to_dict()

        return values

    def originate(self, line_number: int = -1) -> Effect:
        """Originate a new command in the note body with explicit compound medication processing."""
        self._validate_before_effect("originate")

        # Get raw values and apply explicit processing for compound medications
        raw_data = self.values
        processed_data = self._process_compound_medication_fields(raw_data)

        return Effect(
            type=f"ORIGINATE_{self.constantized_key()}_COMMAND",
            payload=json.dumps(
                {
                    "command": self.command_uuid,
                    "note": self.note_uuid,
                    "data": processed_data,
                    "line_number": line_number,
                }
            ),
        )

    def edit(self) -> Effect:
        """Edit the command with explicit compound medication processing."""
        self._validate_before_effect("edit")

        # Get raw values and apply explicit processing for compound medications
        raw_data = self.values
        processed_data = self._process_compound_medication_fields(raw_data)

        return Effect(
            type=f"EDIT_{self.constantized_key()}_COMMAND",
            payload=json.dumps(
                {
                    "command": self.command_uuid,
                    "data": processed_data,
                }
            ),
        )


__exports__ = (
    "PrescribeCommand",
    "CompoundMedicationData",
    # Not defined here but used in a current plugin
    "ClinicalQuantity",
    "Decimal",
)



import os
from typing import Any

import redis
from redis.backoff import ExponentialBackoff
from redis.exceptions import ConnectionError, TimeoutError
from redis.retry import Retry

REDIS_ENDPOINT = os.getenv("REDIS_ENDPOINT")
CUSTOMER_IDENTIFIER = os.getenv("CUSTOMER_IDENTIFIER")
CHANNEL_SUFFIX = "plugin-logging-stream"


class PubSubBase:
    """Base class for pub/sub."""

    def __init__(self) -> None:
        self.redis_endpoint = REDIS_ENDPOINT
        self.channel = self._get_channel_name()
        self.client = self._create_client()

    def _get_channel_name(self) -> str | None:
        if CUSTOMER_IDENTIFIER:
            return f"{CUSTOMER_IDENTIFIER}:{CHANNEL_SUFFIX}"

        return None

    def _create_client(self) -> redis.Redis | None:
        if self.redis_endpoint and self.channel:
            return redis.Redis.from_url(
                self.redis_endpoint,
                decode_responses=True,
                retry=Retry(backoff=ExponentialBackoff(), retries=10),
                retry_on_error=[ConnectionError, TimeoutError, ConnectionResetError],
                health_check_interval=1,
            )

        return None


class Publisher(PubSubBase):
    """Publisher class for pub/sub."""

    def publish(self, message: Any) -> None:
        """Publishes a message to the channel."""
        if self.client and self.channel:
            self.client.publish(self.channel, message)



import datetime

import factory
from django.utils import timezone
from factory.fuzzy import FuzzyDate, FuzzyDateTime

from canvas_sdk.test_utils.factories.patient import PatientFactory
from canvas_sdk.v1.data.medication_history import (
    MedicationHistoryMedication,
    MedicationHistoryMedicationCoding,
    MedicationHistoryResponse,
    MedicationHistoryResponseStatus,
)


class MedicationHistoryMedicationFactory(
    factory.django.DjangoModelFactory[MedicationHistoryMedication]
):
    """Factory for creating a MedicationHistoryMedication."""

    class Meta:
        model = MedicationHistoryMedication

    patient = factory.SubFactory(PatientFactory)
    last_fill_date = FuzzyDateTime(
        start_dt=timezone.now() - datetime.timedelta(days=365), end_dt=timezone.now()
    )

    drug_description = factory.Faker("word")
    strength_value = "10"
    strength_form = "tablet"
    strength_unit_of_measure = "mg"
    quantity = 30.0
    quantity_unit_of_measure = "tablets"
    days_supply = 30
    sig = factory.Faker("sentence", nb_words=4)


class MedicationHistoryMedicationCodingFactory(
    factory.django.DjangoModelFactory[MedicationHistoryMedicationCoding]
):
    """Factory for creating a MedicationHistoryMedicationCoding."""

    class Meta:
        model = MedicationHistoryMedicationCoding

    medication = factory.SubFactory(MedicationHistoryMedicationFactory)

    system = "http://www.nlm.nih.gov/research/umls/rxnorm"
    version = "1.0"
    code = factory.Faker("numerify", text="######")
    display = factory.Faker("word")
    user_selected = False


class MedicationHistoryResponseFactory(
    factory.django.DjangoModelFactory[MedicationHistoryResponse]
):
    """Factory for creating a MedicationHistoryResponse."""

    class Meta:
        model = MedicationHistoryResponse

    patient = factory.SubFactory(PatientFactory)
    status = MedicationHistoryResponseStatus.STATUS_APPROVED
    start_date = FuzzyDate(
        start_date=datetime.date.today() - datetime.timedelta(days=365),
        end_date=datetime.date.today(),
    )
    end_date = FuzzyDate(
        start_date=datetime.date.today(),
        end_date=datetime.date.today() + datetime.timedelta(days=365),
    )
    message_id = factory.Faker("uuid4")
    related_to_message_id = factory.Faker("uuid4")



from datetime import date, datetime
from enum import StrEnum

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand


class GoalCommand(_BaseCommand):
    """A class for managing a Goal command within a specific note."""

    class Meta:
        key = "goal"

    class Priority(StrEnum):
        HIGH = "high-priority"
        MEDIUM = "medium-priority"
        LOW = "low-priority"

    class AchievementStatus(StrEnum):
        IN_PROGRESS = "in-progress"
        IMPROVING = "improving"
        WORSENING = "worsening"
        NO_CHANGE = "no-change"
        ACHIEVED = "achieved"
        SUSTAINING = "sustaining"
        NOT_ACHIEVED = "not-achieved"
        NO_PROGRESS = "no-progress"
        NOT_ATTAINABLE = "not-attainable"

    goal_statement: str = ""
    start_date: date = Field(default_factory=lambda: datetime.now().date())
    due_date: date | None = None
    achievement_status: AchievementStatus | None = None
    priority: Priority | None = None
    progress: str | None = None


__exports__ = ("GoalCommand",)



from django.db import models

from canvas_sdk.v1.data.base import Model


class InvoiceRecipients(models.TextChoices):
    """Choices for invoice recipients."""

    PATIENT = "patient", "Patient"
    GUARANTOR = "guarantor", "Guarantor"


class InvoiceStatus(models.TextChoices):
    """Choices for invoice status."""

    ACTIVE = "active", "Active"
    ERROR = "error", "Error"
    ARCHIVED = "archived", "Archived"


class InvoiceWorkflow(models.TextChoices):
    """Choices for invoice workflow."""

    AUTOMATED = "automated", "Automated"
    ADHOC = "adhoc", "Adhoc"
    BATCH = "batch", "Batch"


class InvoiceSentMeans(models.TextChoices):
    """Choices for how the invoice was sent."""

    MAIL = "mail", "Mail"
    EMAIL = "e-mail", "E-mail"


class Invoice(Model):
    """Represents a full invoice for a patient."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_invoicefull_001"

    originator = models.ForeignKey(
        "v1.CanvasUser", on_delete=models.PROTECT, related_name="generated_invoices"
    )
    recipient = models.ForeignKey("v1.Patient", on_delete=models.CASCADE, related_name="invoices")
    recipient_type = models.CharField(max_length=10, choices=InvoiceRecipients.choices)
    total_amount = models.DecimalField(max_digits=8, decimal_places=2)
    # invoice_pdf = models.FileField(blank=True, null=True, upload_to=invoice_upload)
    status = models.CharField(max_length=10, choices=InvoiceStatus.choices)
    workflow = models.CharField(max_length=10, choices=InvoiceWorkflow.choices)
    # claims = models.ManyToManyField("v1.Claim", related_name="full_invoices")
    error_message = models.TextField()
    sent_mean = models.CharField(max_length=6, choices=InvoiceSentMeans.choices)


__exports__ = (
    "Invoice",
    "InvoiceRecipients",
    "InvoiceStatus",
    "InvoiceWorkflow",
    "InvoiceSentMeans",
)



from django.db import models

from canvas_sdk.v1.data.base import Model


class Coding(Model):
    """A representation of a coding from a terminology system."""

    class Meta:
        abstract = True

    code = models.CharField(max_length=255, blank=True, default="")
    display = models.CharField(max_length=1000)
    system = models.CharField(max_length=255)
    version = models.CharField(max_length=255, blank=True, default="")
    user_selected = models.BooleanField(default=False)


__exports__ = ()



from django.db import models

from canvas_sdk.v1.data.base import TimestampedModel


class Discount(TimestampedModel):
    """Model to represent a discount applied to a claim or patient posting."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_discount_001"

    name = models.CharField(max_length=500)
    adjustment_group = models.CharField(max_length=3)
    adjustment_code = models.CharField(max_length=3)
    discount = models.DecimalField(max_digits=8, decimal_places=2)


__exports__ = ("Discount",)



import json

from django.db import models

from canvas_sdk.v1.data.base import AuditedModel, IdentifiableModel, TimestampedModel
from canvas_sdk.v1.data.task import Task


class Referral(AuditedModel, IdentifiableModel):
    """Referral."""

    class Meta:
        db_table = "canvas_sdk_data_api_referral_001"

    patient = models.ForeignKey("v1.Patient", on_delete=models.DO_NOTHING)
    note = models.ForeignKey("v1.Note", on_delete=models.DO_NOTHING)
    service_provider = models.ForeignKey(
        "v1.ServiceProvider",
        on_delete=models.CASCADE,
        related_name="referrals",
        null=True,
        blank=True,
    )
    assessments = models.ManyToManyField("v1.Assessment", related_name="referrals", blank=True)
    clinical_question = models.CharField(max_length=50)
    priority = models.CharField(max_length=255)
    include_visit_note = models.BooleanField()
    notes = models.TextField()
    date_referred = models.DateTimeField()
    forwarded = models.BooleanField()
    internal_comment = models.TextField()
    internal_task_comment = models.OneToOneField(
        "v1.TaskComment", on_delete=models.SET_NULL, null=True, related_name="referral"
    )
    ignored = models.BooleanField()

    task_ids = models.CharField(max_length=1024)

    def get_task_objects(self) -> "models.QuerySet[Task]":
        """Convert task IDs to Task objects."""
        if self.task_ids:
            task_ids = (
                json.loads(self.task_ids) if isinstance(self.task_ids, str) else self.task_ids
            )
            return Task.objects.filter(id__in=task_ids)
        return Task.objects.none()

    @property
    def task_list(self) -> list[Task]:
        """Convenience property to get task objects."""
        return list(self.get_task_objects())

    def __str__(self) -> str:
        return f"Referral {self.id}"


class ReferralReport(TimestampedModel, IdentifiableModel):
    """ReferralReport."""

    class Meta:
        db_table = "canvas_sdk_data_api_referralreport_001"

    originator = models.ForeignKey(
        "v1.CanvasUser", on_delete=models.DO_NOTHING, null=True, related_name="+"
    )

    review_mode = models.CharField(max_length=2)
    assigned_by = models.ForeignKey(
        "v1.CanvasUser", on_delete=models.DO_NOTHING, null=True, related_name="+"
    )
    junked = models.BooleanField()
    requires_signature = models.BooleanField()
    assigned_date = models.DateTimeField(null=True)
    team_assigned_date = models.DateTimeField(null=True)
    team = models.ForeignKey("v1.Team", on_delete=models.DO_NOTHING, null=True)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="referral_reports"
    )
    referral = models.ForeignKey(
        Referral, on_delete=models.DO_NOTHING, related_name="reports", null=True
    )
    specialty = models.CharField(max_length=250)
    original_date = models.DateField(null=True)
    comment = models.TextField()
    priority = models.BooleanField(default=False)


__exports__ = ("Referral", "ReferralReport")



from typing import Any

from canvas_sdk.effects.base import EffectType, _BaseEffect


class RemoveBillingLineItem(_BaseEffect):
    """
    An Effect that will remove a billing line item in a note footer.
    """

    class Meta:
        effect_type = EffectType.REMOVE_BILLING_LINE_ITEM
        apply_required_fields = ("billing_line_item_id",)

    billing_line_item_id: str | None = None

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {
            "billing_line_item_id": self.billing_line_item_id,
        }


__exports__ = ("RemoveBillingLineItem",)



from enum import Enum

from canvas_sdk.commands.base import _BaseCommand


class ChartSectionReviewCommand(_BaseCommand):
    """A class for managing a Chart Section Review command within a specific note."""

    class Meta:
        key = "chartSectionReview"

    class Sections(Enum):
        CONDITIONS = "conditions"
        SURGICAL_HISTORY = "surgical_history"
        MEDICATIONS = "medications"
        FAMILY_HISTORY = "family_histories"
        ALLERGIES = "allergies"
        IMMUNIZATIONS = "immunizations"

    section: Sections


__exports__ = ("ChartSectionReviewCommand",)



from enum import Enum
from typing import Any

from pydantic import Field
from typing_extensions import TypedDict

from canvas_sdk.effects.base import EffectType, _BaseEffect


class PatientProfileConfiguration(_BaseEffect):
    """
    An Effect that will decide which sections appear, and if they start expanded,
    on the patient's profile in Canvas.
    """

    class Meta:
        effect_type = EffectType.SHOW_PATIENT_PROFILE_SECTIONS

    class Section(Enum):
        DEMOGRAPHICS = "demographics"
        PORTAL = "portal"
        PREFERENCES = "preferences"
        PREFERRED_PHARMACIES = "preferred_pharmacies"
        PATIENT_CONSENTS = "patient_consents"
        CARE_TEAM = "care_team"
        PARENT_GUARDIAN = "parent_guardian"
        ADDRESSES = "addresses"
        TELECOM = "telecom"
        CONTACTS = "contacts"

    class Payload(TypedDict):
        type: "PatientProfileConfiguration.Section"
        start_expanded: bool

    sections: list[Payload] = Field(min_length=1)

    @property
    def values(self) -> dict[str, Any]:
        """The PatientProfileConfiguration's values."""
        return {"sections": [{**s, "type": s["type"].value} for s in self.sections]}

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {"data": self.values}


__exports__ = ("PatientProfileConfiguration",)



from canvas_cli.apps.plugin.plugin import (
    disable,
    enable,
    init,
    install,
    list,
    list_secrets,
    set_secrets,
    uninstall,
    validate_manifest,
)

__all__ = (
    "list_secrets",
    "set_secrets",
    "disable",
    "enable",
    "init",
    "install",
    "list",
    "uninstall",
    "validate_manifest",
)



import json
import os
import shutil
import tarfile
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import TypedDict, cast
from urllib import parse

import psycopg
import requests
import sentry_sdk
from psycopg import Connection
from psycopg.rows import dict_row

from logger import log
from plugin_runner.aws_headers import aws_sig_v4_headers
from plugin_runner.exceptions import (
    InvalidPluginFormat,
    PluginInstallationError,
    PluginUninstallationError,
)
from settings import (
    AWS_ACCESS_KEY_ID,
    AWS_REGION,
    AWS_SECRET_ACCESS_KEY,
    CUSTOMER_IDENTIFIER,
    MEDIA_S3_BUCKET_NAME,
    PLUGIN_DIRECTORY,
    SECRETS_FILE_NAME,
)

# Plugin "packages" include this prefix in the database record for the plugin and the S3 bucket key.
UPLOAD_TO_PREFIX = "plugins"


def open_database_connection() -> Connection:
    """Opens a psycopg connection to the home-app database.

    When running within Aptible, use the database URL, otherwise pull from
    the environment variables.
    """
    if os.getenv("DATABASE_URL"):
        parsed_url = parse.urlparse(os.getenv("DATABASE_URL"))

        return psycopg.connect(
            dbname=cast(str, parsed_url.path[1:]),
            user=cast(str, parsed_url.username),
            password=cast(str, parsed_url.password),
            host=cast(str, parsed_url.hostname),
            port=parsed_url.port,
        )

    APP_NAME = os.getenv("APP_NAME")

    return psycopg.connect(
        dbname=APP_NAME,
        user=os.getenv("DB_USERNAME", "app"),
        password=os.getenv("DB_PASSWORD", "app"),
        host=os.getenv("DB_HOST", f"{APP_NAME}-db"),
        port=os.getenv("DB_PORT", "5432"),
    )


class PluginAttributes(TypedDict):
    """Attributes of a plugin."""

    version: str
    package: str
    secrets: dict[str, str]


def enabled_plugins(plugin_names: list[str] | None = None) -> dict[str, PluginAttributes]:
    """Returns a dictionary of enabled plugins and their attributes.

    If `plugin_names` is provided, only returns those plugins (if enabled).
    """
    conn = open_database_connection()

    with conn.cursor(row_factory=dict_row) as cursor:
        base_query = (
            "SELECT name, package, version, key, value "
            "FROM plugin_io_plugin p "
            "LEFT JOIN plugin_io_pluginsecret s ON p.id = s.plugin_id "
            "WHERE is_enabled"
        )

        params = []
        if plugin_names:
            placeholders = ",".join(["%s"] * len(plugin_names))
            base_query += f" AND name IN ({placeholders})"
            params.extend(plugin_names)

        cursor.execute(base_query, params)
        rows = cursor.fetchall()
        plugins = _extract_rows_to_dict(rows)

    return plugins


def _extract_rows_to_dict(rows: list) -> dict[str, PluginAttributes]:
    plugins = {}

    for row in rows:
        if row["name"] not in plugins:
            plugins[row["name"]] = PluginAttributes(
                version=row["version"],
                package=row["package"],
                secrets={row["key"]: row["value"]} if row["key"] else {},
            )
        else:
            plugins[row["name"]]["secrets"][row["key"]] = row["value"]

    return plugins


@contextmanager
def download_plugin(plugin_package: str) -> Generator[Path, None, None]:
    """Download the plugin package from the S3 bucket."""
    method = "GET"
    host = f"s3-{AWS_REGION}.amazonaws.com"
    bucket = MEDIA_S3_BUCKET_NAME
    customer_identifier = CUSTOMER_IDENTIFIER
    path = f"/{bucket}/{customer_identifier}/{plugin_package}"
    payload = b"This is required for the AWS headers because it is part of the signature"
    pre_auth_headers: dict[str, str] = {}
    query: dict[str, str] = {}
    headers = aws_sig_v4_headers(
        AWS_ACCESS_KEY_ID,
        AWS_SECRET_ACCESS_KEY,
        pre_auth_headers,
        "s3",
        AWS_REGION,
        host,
        method,
        path,
        query,
        payload,
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        prefix_dir = Path(temp_dir) / UPLOAD_TO_PREFIX
        prefix_dir.mkdir()  # create an intermediate directory reflecting the prefix
        download_path = Path(temp_dir) / plugin_package

        with open(download_path, "wb") as download_file:
            response = requests.request(method=method, url=f"https://{host}{path}", headers=headers)
            response.raise_for_status()

            download_file.write(response.content)

        yield download_path


def install_plugin(plugin_name: str, attributes: PluginAttributes) -> None:
    """Install the given Plugin's package into the runtime."""
    try:
        log.info(f'Installing plugin "{plugin_name}", version {attributes["version"]}')

        plugin_installation_path = Path(PLUGIN_DIRECTORY) / plugin_name

        # if plugin exists, first uninstall it
        if plugin_installation_path.exists():
            uninstall_plugin(plugin_name)

        with download_plugin(attributes["package"]) as plugin_file_path:
            extract_plugin(plugin_file_path, plugin_installation_path)

        install_plugin_secrets(plugin_name=plugin_name, secrets=attributes["secrets"])
    except Exception as e:
        log.error(f'Failed to install plugin "{plugin_name}", version {attributes["version"]}: {e}')

        sentry_sdk.capture_exception(e)

        raise PluginInstallationError() from e


def extract_plugin(plugin_file_path: Path, plugin_installation_path: Path) -> None:
    """Extract plugin in `file` to the given `path`."""
    log.info(f'Extracting plugin at "{plugin_file_path}"')

    archive: tarfile.TarFile | None = None

    try:
        if tarfile.is_tarfile(plugin_file_path):
            try:
                with open(plugin_file_path, "rb") as file:
                    archive = tarfile.TarFile.open(fileobj=file)
                    archive.extractall(plugin_installation_path, filter="data")
            except tarfile.ReadError as e:
                log.error(f"Unreadable tar archive: '{plugin_file_path}'")
                sentry_sdk.capture_exception(e)

                raise InvalidPluginFormat from e
        else:
            log.error(f"Unsupported file format: '{plugin_file_path}'")
            raise InvalidPluginFormat
    finally:
        if archive:
            archive.close()


def install_plugin_secrets(plugin_name: str, secrets: dict[str, str]) -> None:
    """Write the plugin's secrets to disk in the package's directory."""
    log.info(f"Writing plugin secrets for '{plugin_name}'")

    secrets_path = Path(PLUGIN_DIRECTORY) / plugin_name / SECRETS_FILE_NAME

    # Did the plugin ship a secrets.json? TOO BAD, IT'S GONE NOW.
    if Path(secrets_path).exists():
        os.remove(secrets_path)

    with open(str(secrets_path), "w") as f:
        json.dump(secrets, f)


def disable_plugin(plugin_name: str) -> None:
    """Disable the given plugin."""
    conn = open_database_connection()
    conn.cursor().execute(
        "UPDATE plugin_io_plugin SET is_enabled = false WHERE name = %s", (plugin_name,)
    )
    conn.commit()
    conn.close()

    uninstall_plugin(plugin_name)


def uninstall_plugin(plugin_name: str) -> None:
    """Remove the plugin from the filesystem."""
    try:
        log.info(f'Uninstalling plugin "{plugin_name}"')

        plugin_path = Path(PLUGIN_DIRECTORY) / plugin_name

        if plugin_path.exists():
            shutil.rmtree(plugin_path)
    except Exception as e:
        raise PluginUninstallationError() from e


def install_plugins() -> None:
    """Install all enabled plugins."""
    log.info("Installing plugins")
    try:
        plugins_dir = Path(PLUGIN_DIRECTORY).resolve()

        if plugins_dir.exists():
            shutil.rmtree(plugins_dir.as_posix())

        plugins_dir.mkdir(parents=False, exist_ok=True)
    except Exception as e:
        raise PluginInstallationError(
            f'Failed to reset plugin directory "{PLUGIN_DIRECTORY}": {e}"'
        ) from e

    for plugin_name, attributes in enabled_plugins().items():
        try:
            install_plugin(plugin_name, attributes)
        except PluginInstallationError as e:
            disable_plugin(plugin_name)

            log.error(
                f'Installation failed for plugin "{plugin_name}", version {attributes["version"]};'
                " the plugin has been disabled"
            )

            sentry_sdk.capture_exception(e)

            continue

    return None



from typing import Any, Generic, TypeVar

T = TypeVar("T")


class WriteOnceProperty(Generic[T]):
    """A descriptor for write-once values."""

    def __set_name__(self, owner: Any, name: str) -> None:
        """Set the name of the property so that the descriptor can access it."""
        self.public_name = name
        self.private_name = f"_{name}"

    def __get__(self, obj: Any, objtype: Any = None) -> T:
        """Retrieve the value of the property."""
        return getattr(obj, self.private_name)

    def __set__(self, obj: Any, value: T) -> None:
        """Set the value of the property.

        Throws an AttributeError if the property already has a value.
        """
        if hasattr(obj, self.private_name):
            raise AttributeError(f"{self.public_name} cannot be changed.")
        setattr(obj, self.private_name, value)


__exports__ = ()



from typing import Any
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_sdk.effects.base import EffectType, _BaseEffect
from canvas_sdk.v1.data import PatientContactPoint


class SendContactVerificationEffect(_BaseEffect):
    """
    An Effect that will send a verification for a Patient Contact Point.
    """

    class Meta:
        effect_type = EffectType.PATIENT_PORTAL__SEND_CONTACT_VERIFICATION

    contact_point_id: str | UUID

    @property
    def values(self) -> dict[str, Any]:
        """The contact point id."""
        return {"contact_point_id": self.contact_point_id}

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        contact_point_exists = PatientContactPoint.objects.filter(id=self.contact_point_id).exists()

        if not contact_point_exists:
            errors.append(
                self._create_error_detail(
                    "value",
                    "Patient Contact Point does not exist",
                    self.contact_point_id,
                )
            )

        return errors


__exports__ = ("SendContactVerificationEffect",)



from typing import cast

from django.db import models

from canvas_sdk.v1.data.base import (
    BaseModelManager,
    CommittableQuerySetMixin,
    ForPatientQuerySetMixin,
    IdentifiableModel,
    TimestampedModel,
    ValueSetLookupQuerySet,
)
from canvas_sdk.v1.data.coding import Coding


class AllergyIntoleranceQuerySet(
    ValueSetLookupQuerySet,
    CommittableQuerySetMixin,
    ForPatientQuerySetMixin,
):
    """AllergyIntoleranceQuerySet."""

    pass


AllergyIntoleranceManager = BaseModelManager.from_queryset(AllergyIntoleranceQuerySet)


class AllergyIntolerance(TimestampedModel, IdentifiableModel):
    """AllergyIntolerance."""

    class Meta:
        db_table = "canvas_sdk_data_api_allergyintolerance_001"

    objects = cast(AllergyIntoleranceQuerySet, AllergyIntoleranceManager())

    deleted = models.BooleanField()
    committer = models.ForeignKey(
        "v1.CanvasUser", on_delete=models.DO_NOTHING, null=True, related_name="+"
    )
    entered_in_error = models.ForeignKey(
        "v1.CanvasUser", on_delete=models.DO_NOTHING, null=True, related_name="+"
    )
    patient = models.ForeignKey(
        "v1.Patient",
        on_delete=models.DO_NOTHING,
        related_name="allergy_intolerances",
        null=True,
    )
    note_id = models.BigIntegerField()
    allergy_intolerance_type = models.CharField(max_length=1)
    category = models.IntegerField()
    status = models.CharField(max_length=20)
    severity = models.CharField(max_length=20)
    onset_date = models.DateField()
    onset_date_original_input = models.CharField(max_length=255)
    last_occurrence = models.DateField()
    last_occurrence_original_input = models.CharField(max_length=255)
    recorded_date = models.DateTimeField()
    narrative = models.CharField(max_length=512)


class AllergyIntoleranceCoding(Coding):
    """AllergyIntoleranceCoding."""

    class Meta:
        db_table = "canvas_sdk_data_api_allergyintolerancecoding_001"

    allergy_intolerance = models.ForeignKey(
        AllergyIntolerance,
        on_delete=models.DO_NOTHING,
        related_name="codings",
        null=True,
    )


__exports__ = ("AllergyIntolerance", "AllergyIntoleranceCoding")



from __future__ import annotations

import ast
import builtins
import importlib
import json
import operator
import sys
import types
from _ast import AnnAssign
from collections.abc import Iterable, Sequence
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypedDict, cast

from frozendict import frozendict
from RestrictedPython import (
    CompileResult,
    PrintCollector,
    RestrictingNodeTransformer,
    compile_restricted_exec,
    safe_builtins,
    utility_builtins,
)
from RestrictedPython.Guards import (
    guarded_iter_unpack_sequence,
    guarded_unpack_sequence,
)
from RestrictedPython.transformer import (
    ALLOWED_FUNC_NAMES,
    FORBIDDEN_FUNC_NAMES,
    INSPECT_ATTRIBUTES,
    copy_locations,
)

if TYPE_CHECKING:

    class ImportedNames(TypedDict):
        """
        Type the stored imported_names dicitionary for mypy.
        """

        names: list[str]
        names_to_module: dict[str, str]


try:
    allowed_module_imports_path = Path(__file__).parent / "allowed-module-imports.json"

    CANVAS_MODULES: dict[str, Iterable[str]] = json.loads(allowed_module_imports_path.read_text())

    for key in CANVAS_MODULES:
        CANVAS_MODULES[key] = set(CANVAS_MODULES[key])

except FileNotFoundError:
    print("Error: Unable to load plugin_runner/allowed-module-imports.json, aborting")
    sys.exit(1)


SAFE_INTERNAL_DUNDER_READ_ATTRIBUTES = {
    "__class__",
    "__dict__",
    "__eq__",
    "__init__",
    "__name__",
}


SAFE_EXTERNAL_DUNDER_READ_ATTRIBUTES = {
    "__dict__",
    "__eq__",
    "__init__",
    "__name__",
}

STANDARD_LIBRARY_MODULES = {
    "__future__": {
        "annotations",
    },
    "_strptime": set(),  # gets imported at runtime via datetime.datetime.strptime()
    "abc": {
        "ABC",
        "abstractmethod",
    },
    "base64": {
        "b64decode",
        "b64encode",
    },
    "collections": {
        "Counter",
        "defaultdict",
    },
    "datetime": {
        "date",
        "datetime",
        "timedelta",
        "timezone",
        "UTC",
    },
    "dateutil": {
        "relativedelta",
    },
    "dateutil.relativedelta": {
        "relativedelta",
    },
    "decimal": {
        "Decimal",
    },
    "enum": {
        "Enum",
        "StrEnum",
    },
    "functools": {
        "reduce",
    },
    "hashlib": {
        "sha256",
    },
    "hmac": {
        "compare_digest",
        "new",
    },
    "http": {
        "HTTPStatus",
    },
    "json": {
        "dumps",
        "loads",
    },
    "operator": {
        "and_",
    },
    "random": {
        "choices",
        "uniform",
        "randint",
    },
    "re": {
        "compile",
        "DOTALL",
        "IGNORECASE",
        "match",
        "search",
        "split",
        "sub",
    },
    "string": {
        "ascii_lowercase",
        "digits",
    },
    "time": {
        "time",
        "sleep",
    },
    "typing": {
        "Any",
        "cast",
        "Dict",
        "Final",
        "Iterable",
        "List",
        "Literal",
        "NamedTuple",
        "NotRequired",
        "Optional",
        "Protocol",
        "Sequence",
        "Tuple",
        "Type",
        "TypedDict",
        "TypeGuard",
        "Union",
    },
    "urllib": {
        "parse",
    },
    "urllib.parse": {
        "urlencode",
        "quote",
    },
    "uuid": {
        "uuid4",
        "UUID",
    },
    "zoneinfo": {
        "ZoneInfo",
    },
}


THIRD_PARTY_MODULES = {
    "arrow": {
        "get",
        "now",
        "utcnow",
    },
    "django.db.models": {
        "BigIntegerField",
        "Case",
        "CharField",
        "Count",
        "F",
        "IntegerField",
        "Model",  # remove when hyperscribe no longer needs it
        "Q",
        "Value",
        "When",
    },
    "django.db.models.expressions": {
        "Case",
        "Value",
        "When",
    },
    "django.db.models.functions": {
        "Coalesce",
        "Trim",
    },
    "django.db.models.query": {
        "QuerySet",
    },
    "django.utils.functional": {
        "cached_property",
    },
    "jwt": {
        "decode",
        "encode",
        "PyJWKClient",
        "ExpiredSignatureError",
        "InvalidTokenError",
    },
    "pydantic": {
        "BaseModel",
        "ConfigDict",
        "conint",
        "constr",
        "Field",
        "RootModel",
        "ValidationError",
    },
    "rapidfuzz": {
        "fuzz",
        "process",
        "utils",
    },
    "requests": {
        "delete",
        "get",
        "patch",
        "post",
        "put",
        "request",
        "RequestException",
        "Response",
    },
}


# The modules in this list are the only ones that can be imported in a sandboxed runtime.
ALLOWED_MODULES = frozendict(
    {
        **CANVAS_MODULES,
        **STANDARD_LIBRARY_MODULES,
        **THIRD_PARTY_MODULES,
    }
)


def _is_known_module(name: str) -> bool:
    return name in ALLOWED_MODULES


def _unrestricted(_ob: Any, *args: Any, **kwargs: Any) -> Any:
    """Return the given object, unmodified."""
    return _ob


def _inplacevar(op: Any, val: Any, expr: Any) -> Any:
    """
    Apply the specified operation to the value and expression.

    NOTE: Does not yet support += concatenation for sequences.
    """
    ops = {
        "-=": operator.isub,
        "@=": operator.imatmul,
        "**=": operator.ipow,
        "*=": operator.imul,
        "//=": operator.ifloordiv,
        "/=": operator.itruediv,
        "&=": operator.iand,
        "%=": operator.imod,
        "^=": operator.ixor,
        "+=": operator.iadd,
        "<<=": operator.ilshift,
        ">>=": operator.irshift,
        "|=": operator.ior,
    }

    if op not in ops:
        raise ValueError(f"Invalid inplace operation: {op}")

    return ops[op](val, expr)


def _apply(_ob: Any, *args: Any, **kwargs: Any) -> Any:
    """Call the bound method with args, support calling super().__init__()."""
    return _ob(*args, **kwargs)


def _find_folder_in_path(file_path: Path, target_folder_name: str) -> Path | None:
    """
    Recursively search for a folder with the specified name in the hierarchy of the given file path.
    """
    file_path = file_path.resolve()

    if file_path.name == target_folder_name:
        return file_path

    # If we've reached the root of the file system, return None
    if file_path.parent == file_path:
        return None

    return _find_folder_in_path(file_path.parent, target_folder_name)


def node_name(node: ast.AST) -> str:
    """
    Given an AST node, return its name.
    """
    if isinstance(node, ast.Call):
        return ".".join(node_name(arg) for arg in node.args)

    if isinstance(node, ast.Constant):
        return str(node.value)

    if isinstance(node, ast.Name):
        return str(node.id)

    return "__unknown__"


class Sandbox:
    """A restricted sandbox for safely executing arbitrary Python code."""

    source_code: str
    namespace: str

    class Transformer(RestrictingNodeTransformer):
        """A node transformer for customizing the sandbox compiler."""

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)

            # we can't just add a self attribute here so we abuse used_names
            # which gets returned as part of the CompileResult
            self.used_names["__imported_names__"] = {
                "names": [],
                "names_to_module": {},
            }

        def handle_names(self, node: ast.Import | ast.ImportFrom) -> None:
            """
            Store imported names.
            """
            module = node.module if isinstance(node, ast.ImportFrom) else None

            for name in node.names:
                name_string = name.asname if name.asname else name.name

                self.used_names["__imported_names__"]["names"].append(name_string)

                if module:
                    self.used_names["__imported_names__"]["names_to_module"][name_string] = module

        def visit_Import(self, node: ast.Import) -> ast.Import:
            """
            Store imported names.
            """
            node = super().visit_Import(node)

            self.handle_names(node)

            return node

        def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom:
            """
            Store imported names.
            """
            node = super().visit_ImportFrom(node)

            self.handle_names(node)

            return node

        def visit_AnnAssign(self, node: AnnAssign) -> AnnAssign:
            """Allow type annotations."""
            return self.node_contents_visit(node)

        def visit_Constant(self, node: ast.Constant) -> ast.Constant:
            """Allow constants."""
            return self.node_contents_visit(node)

        def visit_Match(self, node: ast.Match) -> ast.Match:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchAs(self, node: ast.MatchAs) -> ast.MatchAs:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchClass(self, node: ast.MatchClass) -> ast.MatchClass:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchMapping(self, node: ast.MatchMapping) -> ast.MatchMapping:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchOr(self, node: ast.MatchOr) -> ast.MatchOr:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchSingleton(self, node: ast.MatchSingleton) -> ast.MatchSingleton:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchSequence(self, node: ast.MatchSequence) -> ast.MatchSequence:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchStar(self, node: ast.MatchStar) -> ast.MatchStar:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_MatchValue(self, node: ast.MatchValue) -> ast.MatchValue:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def visit_match_case(self, node: ast.match_case) -> ast.match_case:
            """Allow `match`."""
            return self.node_contents_visit(node)

        def check_import_names(self, node: ast.ImportFrom) -> ast.AST:
            """Check the names being imported.

            This is a protection against rebinding dunder names like
            _getitem_, _write_ via imports.

            => 'from _a import x' is ok, because '_a' is not added to the scope.
            """
            for name in node.names:
                if "*" in name.name and node.module and not _is_known_module(node.module):
                    self.error(node, '"*" imports are not allowed.')

                self.check_name(node, name.name)

                if name.asname:
                    self.check_name(node, name.asname)

            return self.node_contents_visit(node)

        def check_name(
            self,
            node: ast.ImportFrom,
            name: str | None,
            allow_magic_methods: bool = False,
        ) -> None:
            """Check names if they are allowed.

            If ``allow_magic_methods is True`` names in `ALLOWED_FUNC_NAMES`
            are additionally allowed although their names start with `_`.

            Override to turn errors into warnings for leading underscores.
            """
            if name is None:
                return

            if (
                name.startswith("_")
                and name != "_"
                and not (
                    allow_magic_methods and name in ALLOWED_FUNC_NAMES and node.col_offset != 0
                )
            ):
                self.warn(
                    node,
                    f'"{name}" is an invalid variable name because it starts with "_"',
                )
            elif name.endswith("__roles__"):
                self.error(
                    node,
                    f'"{name}" is an invalid variable name because it ends with "__roles__".',
                )
            elif name in FORBIDDEN_FUNC_NAMES:
                self.error(node, f'"{name}" is a reserved name.')

        def visit_Assign(self, node: ast.Assign) -> ast.AST:
            """Check for forbidden assignments."""
            for target in node.targets:
                if (
                    isinstance(target, ast.Name)
                    and target.id.startswith("__")
                    and target.id != "__all__"
                ):
                    self.error(node, f"Assignments to '{target.id}' are not allowed.")
                elif isinstance(target, ast.Tuple | ast.List):
                    self.check_for_name_in_iterable(target)

            return super().visit_Assign(node)

        def check_for_name_in_iterable(self, iterable_node: ast.Tuple | ast.List) -> None:
            """Check if any element of an iterable is a forbidden assignment."""
            for elt in iterable_node.elts:
                if isinstance(elt, ast.Name) and elt.id.startswith("__") and elt.id != "__all__":
                    self.error(iterable_node, f"Assignments to '{elt.id}' are not allowed.")
                elif isinstance(elt, ast.Tuple | ast.List):
                    self.check_for_name_in_iterable(elt)

        def visit_Attribute(self, node: ast.Attribute) -> ast.AST:
            """Checks and mutates attribute access/assignment.

            'a.b' becomes '_getattr_(a, "b")'
            'a.b = c' becomes '_write_(a).b = c'
            'del a.b' becomes 'del _write_(a).b'

            The _write_ function should return a security proxy.

            Override to turn errors into warnings for leading underscores.
            """
            if node.attr.startswith("_") and node.attr != "_":
                self.warn(
                    node,
                    f'"{node.attr}" is an invalid attribute name because it starts with "_".',
                )

            if node.attr.endswith("__roles__"):
                self.error(
                    node,
                    f'"{node.attr}" is an invalid attribute name because it ends with "__roles__".',
                )

            if isinstance(node.ctx, ast.Load):
                node = self.node_contents_visit(node)
                new_node = ast.Call(
                    func=ast.Name("_getattr_", ast.Load()),
                    args=[node.value, ast.Constant(node.attr)],
                    keywords=[],
                )

                copy_locations(new_node, node)
                return new_node

            elif isinstance(node.ctx, ast.Store | ast.Del):
                node = self.node_contents_visit(node)

                new_value = ast.Call(
                    func=ast.Name("_write_", ast.Load()),
                    args=[
                        node.value,
                        ast.Constant(node_name(node.value)),
                        ast.Constant(node.attr),
                    ],
                    keywords=[],
                )

                copy_locations(new_value, node.value)
                node.value = new_value
                return node

            else:  # pragma: no cover
                # Impossible Case only ctx Load, Store and Del are defined in ast.
                raise NotImplementedError(f"Unknown ctx type: {type(node.ctx)}")

        def visit_Subscript(self, node: ast.Subscript) -> ast.AST:
            """Transforms all kinds of subscripts.

            'foo[bar]' becomes '_getitem_(foo, bar)'
            'foo[:ab]' becomes '_getitem_(foo, slice(None, ab, None))'
            'foo[ab:]' becomes '_getitem_(foo, slice(ab, None, None))'
            'foo[a:b]' becomes '_getitem_(foo, slice(a, b, None))'
            'foo[a:b:c]' becomes '_getitem_(foo, slice(a, b, c))'
            'foo[a, b:c] becomes '_getitem_(foo, (a, slice(b, c, None)))'
            'foo[a] = c' becomes '_write_(foo)[a] = c'
            'del foo[a]' becomes 'del _write_(foo)[a]'

            The _write_ function should return a security proxy.
            """
            node = self.node_contents_visit(node)

            # 'AugStore' and 'AugLoad' are defined in 'Python.asdl' as possible
            # 'expr_context'. However, according to Python/ast.c
            # they are NOT used by the implementation => No need to worry here.
            # Instead ast.c creates 'AugAssign' nodes, which can be visited.
            if isinstance(node.ctx, ast.Load):
                new_node = ast.Call(
                    func=ast.Name("_getitem_", ast.Load()),
                    args=[node.value, self.transform_slice(node.slice)],
                    keywords=[],
                )

                copy_locations(new_node, node)

                return new_node
            elif isinstance(node.ctx, ast.Del | ast.Store):
                new_value = ast.Call(
                    func=ast.Name("_write_", ast.Load()),
                    args=[
                        node.value,
                        ast.Constant(node_name(node.value)),
                        ast.Constant(node_name(node.slice)),
                    ],
                    keywords=[],
                )

                copy_locations(new_value, node)
                node.value = new_value

                return node
            else:  # pragma: no cover
                # Impossible Case only ctx Load, Store and Del are defined in ast.
                raise NotImplementedError(f"Unknown ctx type: {type(node.ctx)}")

    def __init__(
        self,
        source_code: Path,
        namespace: str,
        evaluated_modules: dict[str, bool] | None = None,
    ) -> None:
        self.namespace = namespace or "protocols"
        self.package_name = self.namespace.split(".")[0]

        if not source_code.exists():
            raise FileNotFoundError(f"File not found: {source_code}")

        self.source_code_path = source_code.as_posix()
        self.source_code = source_code.read_text()
        package_path = _find_folder_in_path(source_code, self.package_name)
        self.base_path = package_path.parent if package_path else None
        self._evaluated_modules: dict[str, bool] = evaluated_modules or {}

    @cached_property
    def scope(self) -> dict[str, Any]:
        """Return the scope used for evaluation."""
        return {
            "__builtins__": {
                **safe_builtins.copy(),
                **utility_builtins.copy(),
                "__import__": self._safe_import,
                "all": builtins.all,
                "any": builtins.any,
                "classmethod": builtins.classmethod,
                "dict": builtins.dict,
                "enumerate": builtins.enumerate,
                "filter": builtins.filter,
                "getattr": self._safe_getattr,
                "hasattr": builtins.hasattr,
                "iter": builtins.iter,
                "list": builtins.list,
                "map": builtins.map,
                "max": builtins.max,
                "min": builtins.min,
                "next": builtins.next,
                "property": builtins.property,
                "reversed": builtins.reversed,
                "staticmethod": builtins.staticmethod,
                "super": builtins.super,
                "vars": builtins.vars,
            },
            "__is_plugin__": True,
            "__metaclass__": type,
            "__name__": self.namespace,
            "_apply_": _apply,
            "_getattr_": self._safe_getattr,
            "_getitem_": self._safe_getitem,
            "_getiter_": _unrestricted,
            "_inplacevar_": _inplacevar,
            "_iter_unpack_sequence_": guarded_iter_unpack_sequence,
            "_print_": PrintCollector,
            "_unpack_sequence_": guarded_unpack_sequence,
            "_write_": self._safe_write,
        }

    @cached_property
    def compile_result(self) -> CompileResult:
        """Compile the source code into bytecode."""
        return compile_restricted_exec(
            source=self.source_code,
            policy=self.Transformer,
            filename=self.source_code_path,
        )

    @property
    def imported_names(self) -> ImportedNames:
        """Return the imported names collecting during parsing."""
        return self.compile_result.used_names["__imported_names__"]

    @property
    def errors(self) -> tuple[str, ...]:
        """Return errors encountered when compiling the source code."""
        return cast(tuple[str, ...], self.compile_result.errors)

    @property
    def warnings(self) -> tuple[str, ...]:
        """Return warnings encountered when compiling the source code."""
        return cast(tuple[str, ...], self.compile_result.warnings)

    def _is_known_module(self, name: str) -> bool:
        return _is_known_module(name) or self._same_module(name)

    def _get_module(self, module_name: str) -> Path:
        """Get the module path for the given module name."""
        module_relative_path = module_name.replace(".", "/")
        module = Path(cast(Path, self.base_path) / f"{module_relative_path}.py")

        if not module.exists():
            module = Path(cast(Path, self.base_path) / f"{module_relative_path}/__init__.py")

        return module

    def _evaluate_module(self, module_name: str) -> None:
        """Evaluate the given module in the sandbox.

        If the module to import belongs to the same package as the current module,
        evaluate it inside a sandbox.
        """
        # Skip modules already evaluated
        if not self._same_module(module_name) or module_name in self._evaluated_modules:
            return

        module = self._get_module(module_name)
        self._evaluate_implicit_imports(module)

        # Re-check after evaluating implicit imports to avoid duplicate evaluations.
        if module_name not in self._evaluated_modules:
            Sandbox(
                module,
                namespace=module_name,
                evaluated_modules=self._evaluated_modules,
            ).execute()

            self._evaluated_modules[module_name] = True

        # Reload the module if already imported to ensure the latest version is used.
        if sys.modules.get(module_name):
            importlib.reload(sys.modules[module_name])

    def _evaluate_implicit_imports(self, module: Path) -> None:
        """Evaluate implicit imports in the sandbox."""
        # Determine the parent module to check for implicit imports.
        parent = module.parent.parent if module.name == "__init__.py" else module.parent
        base_path = cast(Path, self.base_path)

        # Skip evaluation if the parent module is outside the base path or
        # already the source code root.
        if not parent.is_relative_to(base_path) or parent == base_path:
            return

        module_name = parent.relative_to(base_path).as_posix().replace("/", ".")
        init_file = parent / "__init__.py"

        if module_name not in self._evaluated_modules:
            if init_file.exists():
                # Mark as evaluated to prevent infinite recursion.
                self._evaluated_modules[module_name] = True

                Sandbox(
                    init_file,
                    namespace=module_name,
                    evaluated_modules=self._evaluated_modules,
                ).execute()
            else:
                # Mark as evaluated even if no init file exists to prevent redundant checks.
                self._evaluated_modules[module_name] = True

        self._evaluate_implicit_imports(parent)

    def _same_module(self, module: str) -> bool:
        """
        Return True if `module` is within the plugin code.
        """
        return bool(self.base_path) and module.split(".")[0] == self.package_name

    def _safe_write(
        self,
        _ob: Any,
        name: str | None = None,
        attribute: str | int | None = None,
    ) -> Any:
        """Check if the given obj belongs to a protected resource."""
        is_module = isinstance(_ob, types.ModuleType)

        if is_module:
            if not self._same_module(_ob.__name__):
                raise AttributeError(f"Forbidden assignment to a module attribute: {_ob.__name__}.")
        elif isinstance(_ob, type):
            full_name = f"{_ob.__module__}.{_ob.__qualname__}"
            module_name = _ob.__module__
        else:
            full_name = f"{_ob.__class__.__module__}.{_ob.__class__.__qualname__}"
            module_name = _ob.__class__.__module__

        if attribute is not None:
            if isinstance(_ob, dict):
                value = dict.get(_ob, attribute)
            elif isinstance(_ob, list | tuple) and isinstance(attribute, int):
                value = _ob.__getitem__(attribute)
            elif isinstance(attribute, str):
                value = getattr(_ob, attribute, None)
            else:
                value = None
        else:
            value = None

        if not self._same_module(module_name) and (
            # deny if it was anything imported
            (name and name.split(".")[0] in self.imported_names["names"])
            # deny if it's anything callable
            or callable(value)
            # deny writes to dictionary underscore keys
            or (isinstance(_ob, dict) and isinstance(attribute, str) and attribute.startswith("_"))
        ):
            raise AttributeError(
                f"Forbidden assignment to a non-module attribute: {full_name} "
                f"at {name}.{attribute}."
            )

        return _ob

    def _safe_getitem(self, ob: Any, index: Any) -> Any:
        """
        Prevent access to several classes of items.
        """
        if isinstance(index, str) and index.startswith("_"):
            raise AttributeError(f'"{index}" is an invalid item name because it starts with "_"')

        return ob[index]

    def _safe_getattr(self, _ob: Any, name: Any, default: Any = None) -> Any:
        """
        Prevent access to several classes of attributes.

        Restricted attribute types:

        1. underscored attributes created outside of the defining namespace
        2. attributes used by the `inspect` module
        3. dunder methods except for those we deem safe
        4. if a __exports__ module property is defined, any
           attribute not in that property's value
        """
        is_module = isinstance(_ob, types.ModuleType)

        if is_module:
            module = _ob.__name__
        elif isinstance(_ob, type):
            module = _ob.__module__.split(".")[0]
        else:
            module = _ob.__class__.__module__.split(".")[0]

        if type(name) is not str:
            raise TypeError("type(name) must be str")

        if name in ("format", "format_map") and (
            isinstance(_ob, str) or (isinstance(_ob, type) and issubclass(_ob, str))
        ):
            raise NotImplementedError(
                "Using the format and format_map methods of `str` is not safe"
            )

        if name in INSPECT_ATTRIBUTES:
            raise AttributeError(f'"{name}" is a restricted name.')

        # Code defined in the Sandbox namespace can access its own underscore variables
        if name.startswith("_"):
            if self._same_module(module):
                if name.startswith("__") and name not in SAFE_INTERNAL_DUNDER_READ_ATTRIBUTES:
                    raise AttributeError(
                        f'"{name}" is an invalid attribute name because it starts with "_"'
                    )
            else:
                # Nothing can read dunder attributes except those on our safe list
                if name not in SAFE_EXTERNAL_DUNDER_READ_ATTRIBUTES:
                    raise AttributeError(
                        f'"{name}" is an invalid attribute name because it starts with "__"'
                    )

        exports = getattr(_ob, "__exports__", None)

        if exports:
            if name not in exports:
                raise AttributeError(f'"{name}" is an invalid attribute name (not in __exports__)')
        elif is_module and (module not in ALLOWED_MODULES or name not in ALLOWED_MODULES[module]):
            raise AttributeError(
                f'"{module}.{name}" is an invalid attribute name (not in ALLOWED_MODULES)'
            )

        return getattr(_ob, name, default)

    def _safe_import(
        self,
        name: str,
        globals: Any = None,
        locals: Any = None,
        fromlist: Sequence[str] = (),
        level: int = 0,
    ) -> Any:
        if not self._same_module(name):
            # Disallow importing anything not explicitly allowed by ALLOWED_MODULES
            if name not in ALLOWED_MODULES:
                raise ImportError(f"{name!r} is not an allowed import.")

            if fromlist is not None:
                for item in fromlist:
                    if item not in ALLOWED_MODULES.get(name, set()):
                        raise ImportError(f"{item!r} is not an allowed import from {name!r}.")

        # evaluate the module in the sandbox if needed
        self._evaluate_module(name)

        imported_module = __import__(name, globals, locals, fromlist, level)

        # If the module is part of the plugin, mark it as a plugin module
        if self._same_module(name):
            imported_module.__is_plugin__ = True  # type: ignore[attr-defined]

        return imported_module

    def execute(self) -> dict:
        """Execute the given code in a restricted sandbox."""
        if self.errors:
            raise RuntimeError(f"Code is invalid: {self.errors}")

        exec(self.compile_result.code, self.scope)

        return self.scope


def sandbox_from_module(base_path: Path, module_name: str) -> Sandbox:
    """Sandbox the code execution."""
    module_path = base_path / str(module_name.replace(".", "/") + ".py")

    if not module_path.exists():
        raise ModuleNotFoundError(f'Could not load module "{module_name}"')

    sandbox = Sandbox(module_path, namespace=module_name)

    return sandbox



from django.contrib.postgres.fields import ArrayField
from django.db import models

from canvas_sdk.v1.data.base import IdentifiableModel, Model, TimestampedModel
from canvas_sdk.v1.data.common import ColorEnum, Origin


class TaskType(models.TextChoices):
    """Choices for task types."""

    TASK = "T", "Task"
    REMINDER = "R", "Reminder"


class EventType(models.TextChoices):
    """Choices for event types."""

    EVENT_CHART_OPEN = "CHART_OPEN", "Chart Open"


class TaskStatus(models.TextChoices):
    """Choices for task statuses."""

    COMPLETED = "COMPLETED", "Completed"
    CLOSED = "CLOSED", "Closed"
    OPEN = "OPEN", "Open"


class TaskLabelModule(models.TextChoices):
    """Choices for task label modules."""

    CLAIMS = "claims", "Claims"
    TASKS = "tasks", "Tasks"
    APPOINTMENTS = "appointments", "Appointments"


class Task(TimestampedModel, IdentifiableModel):
    """Task."""

    class Meta:
        db_table = "canvas_sdk_data_api_task_001"

    creator = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="creator_tasks", null=True
    )
    assignee = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="assignee_tasks", null=True
    )
    # TODO - uncomment when Team model is created
    # team = models.ForeignKey('v1.Team', on_delete=models.DO_NOTHING, related_name="tasks", null=True)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, blank=True, related_name="tasks", null=True
    )
    task_type = models.CharField(choices=TaskType.choices, max_length=1)
    tag = models.CharField(max_length=64)
    title = models.TextField()
    due = models.DateTimeField(null=True)
    due_event = models.CharField(choices=EventType.choices, blank=True, max_length=16)
    status = models.CharField(choices=TaskStatus.choices, max_length=9)


class TaskComment(TimestampedModel, IdentifiableModel):
    """TaskComment."""

    class Meta:
        db_table = "canvas_sdk_data_api_taskcomment_001"

    creator = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="comments", null=True
    )
    task = models.ForeignKey(Task, on_delete=models.DO_NOTHING, related_name="comments", null=True)
    body = models.TextField()


class TaskLabel(IdentifiableModel):
    """TaskLabel."""

    class Meta:
        db_table = "canvas_sdk_data_api_tasklabel_001"

    tasks = models.ManyToManyField(Task, related_name="labels", through="TaskTaskLabel")  # type: ignore[var-annotated]
    appointments = models.ManyToManyField(
        "v1.Appointment", related_name="labels", through="v1.AppointmentLabel"
    )
    position = models.IntegerField()
    color = models.CharField(choices=ColorEnum.choices, max_length=50)
    task_association = ArrayField(models.CharField(choices=Origin.choices, max_length=32))
    name = models.CharField(max_length=255)
    active = models.BooleanField()
    modules = ArrayField(models.CharField(choices=TaskLabelModule.choices, max_length=32))


class TaskTaskLabel(Model):
    """M2M for Task -> TaskLabels."""

    class Meta:
        db_table = "canvas_sdk_data_api_tasktasklabel_001"

    task_label = models.ForeignKey(TaskLabel, on_delete=models.DO_NOTHING, null=True)
    task = models.ForeignKey(Task, on_delete=models.DO_NOTHING, null=True)


class TaskMetadata(IdentifiableModel):
    """TaskMetadata."""

    class Meta:
        db_table = "canvas_sdk_data_api_taskmetadata_001"

    task = models.ForeignKey(Task, on_delete=models.CASCADE, related_name="metadata")
    key = models.CharField(max_length=32)
    value = models.CharField(max_length=255)


__exports__ = (
    "TaskType",
    "EventType",
    "TaskStatus",
    "TaskLabelModule",
    "Task",
    "TaskComment",
    "TaskLabel",
    "TaskTaskLabel",
    "TaskMetadata",
)



from canvas_sdk.commands.commands.adjust_prescription import AdjustPrescriptionCommand
from canvas_sdk.commands.commands.allergy import AllergyCommand
from canvas_sdk.commands.commands.assess import AssessCommand
from canvas_sdk.commands.commands.chart_section_review import ChartSectionReviewCommand
from canvas_sdk.commands.commands.close_goal import CloseGoalCommand
from canvas_sdk.commands.commands.diagnose import DiagnoseCommand
from canvas_sdk.commands.commands.exam import PhysicalExamCommand
from canvas_sdk.commands.commands.family_history import FamilyHistoryCommand
from canvas_sdk.commands.commands.follow_up import FollowUpCommand
from canvas_sdk.commands.commands.goal import GoalCommand
from canvas_sdk.commands.commands.history_present_illness import (
    HistoryOfPresentIllnessCommand,
)
from canvas_sdk.commands.commands.imaging_order import ImagingOrderCommand
from canvas_sdk.commands.commands.instruct import InstructCommand
from canvas_sdk.commands.commands.lab_order import LabOrderCommand
from canvas_sdk.commands.commands.medical_history import MedicalHistoryCommand
from canvas_sdk.commands.commands.medication_statement import MedicationStatementCommand
from canvas_sdk.commands.commands.past_surgical_history import (
    PastSurgicalHistoryCommand,
)
from canvas_sdk.commands.commands.perform import PerformCommand
from canvas_sdk.commands.commands.plan import PlanCommand
from canvas_sdk.commands.commands.prescribe import PrescribeCommand
from canvas_sdk.commands.commands.questionnaire import QuestionnaireCommand
from canvas_sdk.commands.commands.reason_for_visit import ReasonForVisitCommand
from canvas_sdk.commands.commands.refer import ReferCommand
from canvas_sdk.commands.commands.refill import RefillCommand
from canvas_sdk.commands.commands.remove_allergy import RemoveAllergyCommand
from canvas_sdk.commands.commands.resolve_condition import ResolveConditionCommand
from canvas_sdk.commands.commands.review_of_systems import ReviewOfSystemsCommand
from canvas_sdk.commands.commands.stop_medication import StopMedicationCommand
from canvas_sdk.commands.commands.structured_assessment import StructuredAssessmentCommand
from canvas_sdk.commands.commands.task import TaskCommand
from canvas_sdk.commands.commands.update_diagnosis import UpdateDiagnosisCommand
from canvas_sdk.commands.commands.update_goal import UpdateGoalCommand
from canvas_sdk.commands.commands.vitals import VitalsCommand

__all__ = __exports__ = (
    "AdjustPrescriptionCommand",
    "AllergyCommand",
    "AssessCommand",
    "ChartSectionReviewCommand",
    "CloseGoalCommand",
    "DiagnoseCommand",
    "FamilyHistoryCommand",
    "FollowUpCommand",
    "GoalCommand",
    "HistoryOfPresentIllnessCommand",
    "ImagingOrderCommand",
    "InstructCommand",
    "LabOrderCommand",
    "MedicalHistoryCommand",
    "MedicationStatementCommand",
    "PastSurgicalHistoryCommand",
    "PerformCommand",
    "PlanCommand",
    "PrescribeCommand",
    "PhysicalExamCommand",
    "QuestionnaireCommand",
    "ReasonForVisitCommand",
    "ReferCommand",
    "RefillCommand",
    "RemoveAllergyCommand",
    "ResolveConditionCommand",
    "ReviewOfSystemsCommand",
    "StopMedicationCommand",
    "StructuredAssessmentCommand",
    "TaskCommand",
    "UpdateDiagnosisCommand",
    "UpdateGoalCommand",
    "VitalsCommand",
)



import json
from typing import Any
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_generated.messages.effects_pb2 import Effect
from canvas_sdk.effects import _BaseEffect
from canvas_sdk.effects.patient import PatientPreferredPharmacy
from canvas_sdk.v1.data import Patient


class CreatePatientPreferredPharmacies(_BaseEffect):
    """Effect to create a Patient External Identifier record."""

    class Meta:
        effect_type = "CREATE_PATIENT_PREFERRED_PHARMACIES"

    pharmacies: list[PatientPreferredPharmacy]
    patient_id: str | UUID

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if not Patient.objects.filter(id=self.patient_id).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Patient with ID {self.patient_id} does not exist.",
                    self.patient_id,
                )
            )

        default_count = sum(1 for pharmacy in self.pharmacies if pharmacy.default)
        if default_count > 1:
            errors.append(
                self._create_error_detail(
                    "value",
                    "Only one preferred pharmacy can be set as default.",
                    None,
                )
            )

        return errors

    @property
    def values(self) -> dict[str, Any]:
        """Return the values of the preferred pharmacies."""
        return {
            "preferred_pharmacies": [
                preferred_pharmacy.to_dict() for preferred_pharmacy in self.pharmacies
            ],
            "patient_id": str(self.patient_id),
        }

    def create(self) -> Effect:
        """Create Patient Preferred Pharmacies."""
        self._validate_before_effect("create")

        return Effect(
            type=self.Meta.effect_type,
            payload=json.dumps(
                {
                    "data": self.values,
                }
            ),
        )


__exports__ = ("CreatePatientPreferredPharmacies",)



import inspect
import json
import re
import traceback
from abc import ABC
from base64 import b64decode
from collections.abc import Callable
from functools import cached_property
from http import HTTPStatus
from typing import Any, ClassVar, Protocol, TypeVar, cast
from urllib.parse import parse_qsl

import sentry_sdk

from canvas_sdk.effects import Effect, EffectType
from canvas_sdk.effects.simple_api import JSON, JSONResponse, Response
from canvas_sdk.events import Event, EventType
from canvas_sdk.handlers.base import BaseHandler
from logger import log
from plugin_runner.exceptions import PluginError

from .exceptions import AuthenticationError, InvalidCredentialsError
from .security import Credentials
from .tools import CaseInsensitiveMultiDict, MultiDict, separate_headers


class FormPart(Protocol):
    """
    Protocol for representing a form part in the body of a multipart/form-data request.

    A form part can represent a simple string value, or a file with a content type.
    """

    @staticmethod
    def is_file() -> bool:
        """Return True or False depending on whether the form part represents a file."""
        ...


class StringFormPart(FormPart):
    """Class for representing a form part that is a simple string value."""

    def __init__(self, name: str, value: str) -> None:
        self.name = name
        self.value = value

    @staticmethod
    def is_file() -> bool:
        """Return True or False depending on whether the form part represents a file."""
        return False

    def __eq__(self, other: object) -> bool:
        if isinstance(other, FileFormPart):
            return False

        if not isinstance(other, StringFormPart):
            return NotImplemented

        return self.name == other.name and self.value == other.value


class FileFormPart(FormPart):
    """Class for representing a form part that is a file."""

    def __init__(
        self, name: str, filename: str, content: bytes, content_type: str | None = None
    ) -> None:
        self.name = name
        self.filename = filename
        self.content = content
        self.content_type = content_type

    @staticmethod
    def is_file() -> bool:
        """Return True or False depending on whether the form part represents a file."""
        return True

    def __eq__(self, other: object) -> bool:
        if isinstance(other, StringFormPart):
            return False

        if not isinstance(other, FileFormPart):
            return NotImplemented

        return all(
            (
                self.name == other.name,
                self.filename == other.filename,
                self.content == other.content,
                self.content_type == other.content_type,
            )
        )


def parse_multipart_form(form: bytes, boundary: str) -> MultiDict[str, FormPart]:
    """Parse a multipart form and return a dict of string to list of form parts."""
    form_data: list[tuple[str, FormPart]] = []

    # Split the body by the boundary value and iterate over the parts. The first and last
    # parts can be skipped because there are delimiters on at the start and end of the body
    parts = form.split(f"--{boundary}".encode())
    for part in parts[1:-1]:
        # Each part may be either a simple string value or a file. Simple string values
        # will just have a name and a value, whereas files will have a name, content type,
        # filename, and value.
        name = None
        content_type = None
        filename = None

        # Split the part into the headers and the value (i.e. the content)
        value: str | bytes
        headers, value = part.split(b"\r\n\r\n", maxsplit=1)

        # Iterate over the headers and extract the name, filename, and content type
        for header in headers.decode().split("\r\n"):
            # There are only two possible headers: Content-Disposition and Content-Type
            if header.lower().startswith("content-disposition: form-data;"):
                # Iterate over the content disposition parameters to get the form name and
                # filename
                for parameter in header.split(";")[1:]:
                    parameter_name, parameter_value = parameter.strip().split("=")

                    # Strip the quotes from the value
                    parameter_value = parameter_value[1:-1]

                    if parameter_name == "name":
                        name = parameter_value
                    elif parameter_name == "filename":
                        filename = parameter_value
            elif header.lower().startswith("content-type"):
                # Files will have a content type, so grab it
                content_type = header.split(":")[1].strip()

        if not name or not value:
            raise RuntimeError("Invalid multipart/form-data request body")

        # Strip off the trailing newline characters from the value
        value = value[:-2]

        # Now we have all the data, so append it to the list of form data
        if filename:
            # Because a filename was provided, we know it's a file and not a simple string value
            form_data.append(
                (
                    name,
                    FileFormPart(
                        name=name, filename=filename, content=value, content_type=content_type
                    ),
                )
            )
        else:
            # Decode the string before adding it
            form_data.append((name, StringFormPart(name, value.decode())))

    return MultiDict(form_data)


class Request:
    """Request class for incoming requests to the API."""

    def __init__(self, event: Event, path_pattern: re.Pattern) -> None:
        self.method = event.context["method"]
        self.path = event.context["path"]
        self.query_string = event.context["query_string"]
        self._body = event.context["body"]
        self.headers = CaseInsensitiveMultiDict(separate_headers(event.context["headers"]))

        match = path_pattern.fullmatch(event.context["path"])
        self.path_params = match.groupdict() if match else {}

        self.query_params = MultiDict(parse_qsl(self.query_string))

        # Parse the content type and any included content type parameters
        content_type = self.headers.get("Content-Type")
        self._content_type_parameters = {}
        if content_type:
            content_type, *parameters = content_type.split(";")
            self.content_type = content_type.strip()
            for parameter in parameters:
                name, value = parameter.strip().split("=")
                self._content_type_parameters[name] = value
        else:
            self.content_type = None

    @cached_property
    def body(self) -> bytes:
        """Decode and return the response body."""
        return b64decode(self._body)

    def json(self) -> JSON:
        """Return the response body as a JSON dict."""
        return json.loads(self.body)

    def text(self) -> str:
        """Return the response body as plain text."""
        return self.body.decode()

    def form_data(self) -> MultiDict[str, FormPart]:
        """Return the response body as a dict of string to list of FormPart objects."""
        form_data: MultiDict[str, FormPart]

        if self.content_type == "application/x-www-form-urlencoded":
            # For request bodies that are URL-encoded, just parse them and return them as simple
            # form parts
            form_data = MultiDict(
                (name, StringFormPart(name, value)) for name, value in parse_qsl(self.body.decode())
            )
        elif self.content_type == "multipart/form-data":
            # Parse request bodies that are multipart forms
            form_data = parse_multipart_form(self.body, self._content_type_parameters["boundary"])
        else:
            raise RuntimeError(f"Cannot parse content type {self.content_type} as form data")

        return form_data


SimpleAPIType = TypeVar("SimpleAPIType", bound="SimpleAPIBase")

RouteHandler = Callable[[SimpleAPIType], list[Response | Effect]]


def get(path: str) -> Callable[[RouteHandler], RouteHandler]:
    """Decorator for adding API GET routes."""
    return _handler_decorator("GET", path)


def post(path: str) -> Callable[[RouteHandler], RouteHandler]:
    """Decorator for adding API POST routes."""
    return _handler_decorator("POST", path)


def put(path: str) -> Callable[[RouteHandler], RouteHandler]:
    """Decorator for adding API PUT routes."""
    return _handler_decorator("PUT", path)


def delete(path: str) -> Callable[[RouteHandler], RouteHandler]:
    """Decorator for adding API DELETE routes."""
    return _handler_decorator("DELETE", path)


def patch(path: str) -> Callable[[RouteHandler], RouteHandler]:
    """Decorator for adding API PATCH routes."""
    return _handler_decorator("PATCH", path)


def _handler_decorator(method: str, path: str) -> Callable[[RouteHandler], RouteHandler]:
    if not path.startswith("/"):
        raise PluginError(f"Route path '{path}' must start with a forward slash")

    def decorator(handler: RouteHandler) -> RouteHandler:
        """Mark the handler with the HTTP method and path."""
        handler.route = (method, path)  # type: ignore[attr-defined]

        return handler

    return decorator


class SimpleAPIBase(BaseHandler, ABC):
    """Abstract base class for HTTP APIs."""

    RESPONDS_TO = [
        EventType.Name(EventType.SIMPLE_API_AUTHENTICATE),
        EventType.Name(EventType.SIMPLE_API_REQUEST),
    ]

    _ROUTES: ClassVar[dict[str, list[tuple[re.Pattern, RouteHandler]]]]
    _PATH_PARAM_REGEX = re.compile("<([a-zA-Z_][a-zA-Z0-9_]*)>")

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        # Build the registry of routes so that requests can be routed to the correct handler. This
        # is done by iterating over the methods on the class and looking for methods that have been
        # marked by the handler decorators (get, post, etc.).
        cls._ROUTES = {}
        for attr in cls.__dict__.values():
            if callable(attr) and (route := getattr(attr, "route", None)):
                method, relative_path = route
                path = f"{cls._path_prefix()}{relative_path}"

                # Convert the path to a regular expression pattern so that any path parameters can
                # be extracted later
                try:
                    path_pattern = re.compile(path.replace("<", "(?P<").replace(">", ">[^/]+)"))
                except re.error as error:
                    raise PluginError(
                        f"Path parameter names in route '{path}' must be unique"
                    ) from error

                cls._ROUTES.setdefault(method, [])
                cls._ROUTES[method].append((path_pattern, attr))

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        # Determine the first handler that matches the path based on the path pattern
        self._path_pattern = None
        self._handler = None
        for path_pattern, handler in self._ROUTES.get(self.event.context["method"], ()):
            if path_pattern.fullmatch(self.event.context["path"]):
                self._path_pattern = path_pattern
                self._handler = handler
                break

    @classmethod
    def _path_prefix(cls) -> str:
        return ""

    @cached_property
    def request(self) -> Request:
        """Return the request object from the event."""
        return Request(self.event, cast(re.Pattern, self._path_pattern))

    def compute(self) -> list[Effect]:
        """Handle the authenticate or request event."""
        try:
            if self.event.type == EventType.SIMPLE_API_AUTHENTICATE:
                return self._authenticate()
            elif self.event.type == EventType.SIMPLE_API_REQUEST:
                return self._handle_request()
            else:
                raise AssertionError(f"Cannot handle event type {EventType.Name(self.event.type)}")
        except Exception as exception:
            for error_line_with_newlines in traceback.format_exception(exception):
                for error_line in error_line_with_newlines.split("\n"):
                    log.error(error_line)

            sentry_sdk.capture_exception(exception)

            return [Response(status_code=HTTPStatus.INTERNAL_SERVER_ERROR).apply()]

    def _authenticate(self) -> list[Effect]:
        """Authenticate the request."""
        try:
            # Create the credentials object
            credentials_cls = inspect.get_annotations(self.authenticate, eval_str=True).get(
                "credentials"
            )

            if not credentials_cls or not issubclass(credentials_cls, Credentials):
                raise PluginError(
                    f"Cannot determine authentication scheme for {self.request.path}; "
                    "please specify the type of credentials your endpoint requires"
                )
            credentials = credentials_cls(self.request)

            # Pass the credentials object into the developer-defined authenticate method. If
            # authentication succeeds, return a 200 back to home-app, otherwise return a response
            # with the error
            if self.authenticate(credentials):
                return [Response(status_code=HTTPStatus.OK).apply()]
            else:
                raise InvalidCredentialsError
        except AuthenticationError as error:
            return [
                JSONResponse(
                    content={"error": str(error)}, status_code=HTTPStatus.UNAUTHORIZED
                ).apply()
            ]

    def _handle_request(self) -> list[Effect]:
        """Route the incoming request to the handler method based on the HTTP method and path."""
        # Handle the request
        effects = cast(RouteHandler, self._handler)(self)

        # Iterate over the returned effects and:
        # 1. Change any response objects to response effects
        # 2. Detect if the handler returned multiple responses, and if it did, log an error and
        #    return only a response effect with a 500 Internal Server Error.
        # 3. Detect if the handler returned an error response object, and if it did, return only
        #    the response effect.
        response_found = False
        for index, effect in enumerate(effects):
            # Change the response object to a response effect
            if isinstance(effect, Response):
                effects[index] = effect.apply()

            if cast(Effect, effects[index]).type == EffectType.SIMPLE_API_RESPONSE:
                # If a response has already been found, return an error response immediately
                if response_found:
                    log.error(f"Multiple responses provided by {SimpleAPI.__name__} handler")
                    return [Response(status_code=HTTPStatus.INTERNAL_SERVER_ERROR).apply()]
                else:
                    response_found = True

                # Get the status code of the response. If the initial response was an object and
                # not an effect, get it from the response object to avoid having to deserialize
                # the payload.
                if isinstance(effect, Response):
                    status_code = effect.status_code
                else:
                    status_code = json.loads(effect.payload)["status_code"]

                # If the handler returned an error response, return only that response effect
                # and omit any other included effects
                if 400 <= status_code <= 599:
                    return [cast(Effect, effects[index])]

        return cast(list[Effect], effects)

    def accept_event(self) -> bool:
        """Ignore the event if the handler does not implement the route."""
        return self._handler is not None

    def authenticate(self, credentials: Credentials) -> bool:
        """Method the user should override to authenticate requests."""
        return False


class SimpleAPI(SimpleAPIBase, ABC):
    """Base class for HTTP APIs."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Detect errors the user makes when defining their handler.

        Prevent developers from defining multiple handlers for the same route, and from defining
        methods that clash with base class methods.
        """
        if (prefix := getattr(cls, "PREFIX", None)) and not prefix.startswith("/"):
            raise PluginError(f"Route prefix '{prefix}' must start with a forward slash")

        super().__init_subclass__(**kwargs)

        routes = set()
        route_handler_method_names = set()
        for name, value in cls.__dict__.items():
            if callable(value) and hasattr(value, "route"):
                if value.route in routes:
                    method, path = value.route
                    raise PluginError(
                        f"The route {method} {path} must only be handled by one route handler"
                    )
                routes.add(value.route)
                route_handler_method_names.add(name)

        for superclass in cls.__mro__[1:]:
            if names := route_handler_method_names.intersection(superclass.__dict__):
                raise PluginError(
                    f"{SimpleAPI.__name__} subclass route handler methods are overriding base "
                    f"class attributes: {', '.join(f'{cls.__name__}.{name}' for name in names)}"
                )

    @classmethod
    def _path_prefix(cls) -> str:
        # getattr needs a default else it will raise an exception. We also need to ensure that the
        # final value is a string if the user specifies "None" as the prefix because this value gets
        # prepended to the URL path
        return getattr(cls, "PREFIX", "") or ""


class SimpleAPIRoute(SimpleAPIBase, ABC):
    """Base class for HTTP API routes."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Automatically mark the get, post, put, delete, and patch methods as handler methods."""
        if hasattr(cls, "PREFIX"):
            raise PluginError(
                f"Setting a PREFIX value on a {SimpleAPIRoute.__name__} is not allowed"
            )

        for attr_name, attr_value in cls.__dict__.items():
            decorator: Callable | None
            match attr_name:
                case "get":
                    decorator = get
                case "post":
                    decorator = post
                case "put":
                    decorator = put
                case "delete":
                    decorator = delete
                case "patch":
                    decorator = patch
                case _:
                    decorator = None

            if not callable(attr_value):
                continue

            if hasattr(attr_value, "route"):
                raise PluginError(
                    f"Using the api decorator on subclasses of {SimpleAPIRoute.__name__} is not "
                    "allowed"
                )

            if decorator is None:
                continue

            path = getattr(cls, "PATH", None)
            if not path:
                raise PluginError(f"PATH must be specified on a {SimpleAPIRoute.__name__}")

            decorator(path)(attr_value)

        super().__init_subclass__(**kwargs)

    def get(self) -> list[Response | Effect]:
        """Stub method for GET handler."""
        return []

    def post(self) -> list[Response | Effect]:
        """Stub method for POST handler."""
        return []

    def put(self) -> list[Response | Effect]:
        """Stub method for PUT handler."""
        return []

    def delete(self) -> list[Response | Effect]:
        """Stub method for DELETE handler."""
        return []

    def patch(self) -> list[Response | Effect]:
        """Stub method for PATCH handler."""
        return []


__exports__ = (
    "FormPart",
    "StringFormPart",
    "FileFormPart",
    "parse_multipart_form",
    "Request",
    "SimpleAPIType",
    "RouteHandler",
    "get",
    "post",
    "put",
    "delete",
    "patch",
    "SimpleAPIBase",
    "SimpleAPI",
    "SimpleAPIRoute",
    # Not defined here but used in an existing plugin
    "Credentials",
)



from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class UpdateDiagnosisCommand(BaseCommand):
    """A class for managing an Update Diagnosis command within a specific note."""

    class Meta:
        key = "updateDiagnosis"

    condition_code: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "condition"}
    )
    new_condition_code: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "new_condition"}
    )
    background: str | None = None
    narrative: str | None = None


__exports__ = ("UpdateDiagnosisCommand",)



from django.db import models

from canvas_sdk.v1.data.base import IdentifiableModel, Model, TimestampedModel
from canvas_sdk.v1.data.common import TaxIDType


class PracticeLocationPOS(models.TextChoices):
    """PracticeLocationPOS choices."""

    PHARMACY = "01", "Pharmacy"
    TELEHEALTH = "02", "Telehealth"
    SCHOOL = "03", "Education Facility"
    HOMELESS_SHELTER = "04", "Homeless Shelter"
    PRISON = "09", "Prison"
    TELEHEALTH_IN_PATIENT_HOME = "10", "Telehealth in Patient's Home"
    OFFICE = "11", "Office"
    HOME = "12", "Home"
    ASSISTED_LIVING = "13", "Asssisted Living Facility"
    GROUP_HOME = "14", "Group Home"
    MOBILE = "15", "Mobile Unit"
    WALK_IN_RETAIL = "17", "Walk-In Retail Health Clinic"
    OFF_CAMPUS_OUTPATIENT_HOSPITAL = "19", "Off-Campus Outpatient Hospital"
    URGENT_CARE = "20", "Urgent Care Facility"
    INPATIENT_HOSPITAL = "21", "Inpatient Hospital"
    ON_CAMPUS_OUTPATIENT_HOSPITAL = "22", "On-Campus Outpatient Hospital"
    ER_HOSPITAL = "23", "Emergency Room Hospital"
    AMBULATORY_SURGERY_CENTER = "24", "Ambulatory Surgery Center"
    BIRTHING_CENTER = "25", "Birthing Center"
    MILITARY_FACILITY = "26", "Military Treatment Facility"
    STREET = "27", "Outreach Site / Street"
    SNF = "31", "Skilled Nursing Facility"
    NURSING = "32", "Nursing Facility"
    CUSTODIAL = "33", "Custodial Care Facility"
    HOSPICE = "34", "Hospice"
    AMBULANCE_LAND = "41", "Ambulance Land"
    AMBULANCE_AIR_WATER = "42", "Ambulance Air or Water"
    INDEPENDENT_CLINIC = "49", "Independent Clinic"
    FQHC = "50", "Federally Qualified Health Center"
    PSYCH = "51", "Inpatient Psychiatric Facility"
    PSYCH_PARTIAL = "52", "Inpatient Psychiatric Facility - Partial Hospitalization"
    MENTAL_HEALTH_CENTER = "53", "Community Mental Health Center"
    INTERMEDIATE_MENTAL = "54", "Intermediate Care Facility for Mentally Retarded"
    SUBSTANCE_RESIDENTIAL = "55", "Residential Substance Abuse Treatment Facility"
    PSYCH_RESIDENTIAL = "56", "Psychiatric Residential Treatment Center"
    SUBSTANCE_NON_RESIDENTIAL = "57", "Non-Residential Substance Abuse Treatment Facility"
    MASS_IMMUNIZATION = "60", "Mass Immunization Center"
    INPATIENT_REHAB = "61", "Inpatient Rehabilitation Facility"
    OUTPATIENT_REHAB = "62", "Outpatient Rehabilitation Facility"
    ESRD = "65", "End-Stage Renal Disease Treatment Facility"
    PUBLIC_CLINIC = "71", "State or Local Public Health Clinic"
    RURAL_CLINIC = "72", "Rural Health Clinic"
    INDEPENDENT_LAB = "81", "Independent Laboratory"
    OTHER = "99", "Other Place of Service"


class PracticeLocation(TimestampedModel, IdentifiableModel):
    """PracticeLocation."""

    class Meta:
        db_table = "canvas_sdk_data_api_practicelocation_001"

    organization = models.ForeignKey(
        "v1.Organization", on_delete=models.DO_NOTHING, related_name="practice_locations", null=True
    )
    place_of_service_code = models.CharField(choices=PracticeLocationPOS.choices, max_length=2)
    full_name = models.CharField(max_length=255)
    short_name = models.CharField(max_length=255)
    background_image_url = models.CharField(max_length=255)
    background_gradient = models.CharField(max_length=255)
    active = models.BooleanField(default=True)
    npi_number = models.CharField(max_length=10)
    bill_through_organization = models.BooleanField()
    tax_id = models.CharField(max_length=25)
    tax_id_type = models.CharField(choices=TaxIDType.choices, max_length=1)
    billing_location_name = models.CharField(max_length=255)
    group_npi_number = models.CharField(max_length=10)
    taxonomy_number = models.CharField(max_length=10)
    include_zz_qualifier = models.BooleanField(default=False)

    def __str__(self) -> str:
        return self.full_name


class PracticeLocationSetting(Model):
    """PracticeLocationSetting."""

    class Meta:
        db_table = "canvas_sdk_data_api_practicelocationsetting_001"

    practice_location = models.ForeignKey(
        "v1.PracticeLocation", on_delete=models.DO_NOTHING, related_name="settings", null=True
    )
    name = models.CharField(max_length=100)
    value = models.JSONField()

    def __str__(self) -> str:
        return self.name


__exports__ = (
    "PracticeLocationPOS",
    "PracticeLocation",
    "PracticeLocationSetting",
)



from canvas_sdk.commands.commands.questionnaire import QuestionnaireCommand
from canvas_sdk.commands.commands.questionnaire.toggle_questions import ToggleQuestionsMixin


class PhysicalExamCommand(ToggleQuestionsMixin, QuestionnaireCommand):
    """A class for managing physical exam command."""

    class Meta:
        key = "exam"


__exports__ = ("PhysicalExamCommand",)



from typing import Literal
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_generated.messages.effects_pb2 import EffectType
from canvas_sdk.effects.fax.base import BaseFaxEffect
from canvas_sdk.v1.data import Note


class FaxNoteEffect(BaseFaxEffect):
    """Fax note effect."""

    class Meta:
        effect_type = EffectType.FAX_NOTE

    note_id: str | UUID

    def _get_error_details(self, method: Literal["apply"]) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)
        if self.note_id and not Note.objects.filter(id=self.note_id).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Note with ID {self.note_id} does not exist.",
                    self.note_id,
                )
            )

        return errors

    @property
    def values(self) -> dict[str, str]:
        """Return the values of the note fax effect."""
        values = super().values
        return {
            **values,
            "note_id": str(self.note_id),
        }


__exports__ = ()



from functools import cached_property

from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models.enums import TextChoices
from timezone_utils.fields import TimeZoneField

from canvas_sdk.v1.data.base import IdentifiableModel, Model, TimestampedModel
from canvas_sdk.v1.data.common import (
    AddressState,
    AddressType,
    AddressUse,
    ContactPointState,
    ContactPointSystem,
    ContactPointUse,
    PersonSex,
    TaxIDType,
)
from canvas_sdk.v1.data.utils import create_key


class Staff(TimestampedModel):
    """Staff."""

    class Meta:
        db_table = "canvas_sdk_data_api_staff_001"

    def __str__(self) -> str:
        return f"{self.first_name} {self.last_name}"

    id = models.CharField(
        max_length=32,
        db_column="key",
        unique=True,
        editable=False,
        default=create_key,
    )
    prefix = models.CharField(max_length=100)
    suffix = models.CharField(max_length=100)
    first_name = models.CharField(max_length=255)
    middle_name = models.CharField(max_length=255)
    last_name = models.CharField(max_length=255)
    maiden_name = models.CharField(max_length=255)
    nickname = models.CharField(max_length=255)
    previous_names = models.JSONField(default=list)
    birth_date = models.DateField(null=True)
    sex_at_birth = models.CharField(choices=PersonSex.choices, max_length=3)
    sexual_orientation_term = models.CharField(max_length=255)
    sexual_orientation_code = models.CharField(max_length=255)
    gender_identity_term = models.CharField(max_length=255)
    gender_identity_code = models.CharField(max_length=255)
    preferred_pronouns = models.CharField(max_length=255)
    biological_race_codes = ArrayField(models.CharField(max_length=100))
    biological_race_terms = ArrayField(models.CharField(max_length=255))
    cultural_ethnicity_codes = ArrayField(models.CharField(max_length=100))
    cultural_ethnicity_terms = ArrayField(models.CharField(max_length=255))
    last_known_timezone = TimeZoneField(null=True)
    active = models.BooleanField(default=True)
    primary_practice_location = models.ForeignKey(
        "v1.PracticeLocation", on_delete=models.DO_NOTHING, null=True
    )
    npi_number = models.CharField(max_length=10)
    nadean_number = models.CharField(max_length=20)
    group_npi_number = models.CharField(max_length=10)
    bill_through_organization = models.BooleanField(default=True)
    tax_id = models.CharField(max_length=25)
    tax_id_type = models.CharField(choices=TaxIDType.choices, max_length=1)
    spi_number = models.CharField(max_length=50)
    # TODO - uncomment when Language is developed
    # language = models.ForeignKey('v1.Language', on_delete=models.DO_NOTHING, related_name="staff_speakers", null=True)
    # language_secondary = models.ForeignKey('v1.Language', on_delete=models.DO_NOTHING, related_name="staff_secondary_speakers", null=True)
    personal_meeting_room_link = models.URLField(null=True)
    state = models.JSONField(default=dict, blank=True)
    user = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    schedule_column_ordering = models.IntegerField(default=999)
    default_supervising_provider = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="supervising_team", null=True
    )

    @property
    def photo_url(self) -> str:
        """Return the URL of the staff's first photo."""
        photo = self.photos.first()
        return photo.url if photo else "https://d3hn0m4rbsz438.cloudfront.net/avatar1.png"

    @cached_property
    def full_name(self) -> str:
        """Return Staff's first + last name."""
        return f"{self.first_name} {self.last_name}"

    @cached_property
    def top_clinical_role(self) -> "StaffRole | None":
        """Returns the topmost clinical role to assist in determining privilege levels.

        Returns:
            StaffRole | None: the topmost clinical role of the staff member.
        """
        roles = [
            role
            for role in self.roles.all()
            if role.domain in StaffRole.RoleDomain.clinical_domains()
        ]

        if not roles:
            return None

        return roles[0]

    @cached_property
    def top_role_abbreviation(self) -> str | None:
        """Returns the abbreviation string for the topmost role that the Staff object has.

        Returns:
            Optional[str]: The abbreviation for the topmost role, if available.
        """
        return self.top_clinical_role.public_abbreviation if self.top_clinical_role else None

    @cached_property
    def credentialed_name(self) -> str:
        """Returns the full name of the staff member, suffixed with their topmost credential.

        Returns:
            str: The credentialed full name of the staff member.
        """
        return " ".join(filter(bool, [self.full_name, self.top_role_abbreviation or ""]))


class StaffContactPoint(IdentifiableModel):
    """StaffContactPoint."""

    class Meta:
        db_table = "canvas_sdk_data_api_staffcontactpoint_001"

    system = models.CharField(choices=ContactPointSystem.choices, max_length=20)
    value = models.CharField(max_length=100)
    use = models.CharField(
        choices=ContactPointUse.choices, max_length=20, default=ContactPointUse.HOME
    )
    use_notes = models.CharField(max_length=255)
    rank = models.IntegerField(default=1)
    state = models.CharField(
        choices=ContactPointState.choices, max_length=20, default=ContactPointState.ACTIVE
    )
    staff = models.ForeignKey(Staff, on_delete=models.DO_NOTHING, related_name="telecom")


class StaffAddress(IdentifiableModel):
    """StaffAddress."""

    class Meta:
        db_table = "canvas_sdk_data_api_staffaddress_001"

    line1 = models.CharField(max_length=255, default="", blank=True)
    line2 = models.CharField(max_length=255, default="", blank=True)
    city = models.CharField(max_length=255)
    district = models.CharField(max_length=255, blank=True, default="")
    state_code = models.CharField(max_length=2)
    postal_code = models.CharField(max_length=255)
    use = models.CharField(choices=AddressUse.choices, max_length=10, default=AddressUse.HOME)
    type = models.CharField(choices=AddressType.choices, max_length=10, default=AddressType.BOTH)
    longitude = models.FloatField(null=True, default=None, blank=True)
    latitude = models.FloatField(null=True, default=None, blank=True)
    start = models.DateField(null=True, blank=True)
    end = models.DateField(null=True, blank=True)
    country = models.CharField(max_length=255)
    state = models.CharField(
        choices=AddressState.choices, max_length=20, default=AddressState.ACTIVE
    )
    staff = models.ForeignKey(Staff, on_delete=models.DO_NOTHING, related_name="addresses")


class StaffPhoto(TimestampedModel):
    """StaffPhoto."""

    class Meta:
        db_table = "canvas_sdk_data_api_staffphoto_001"

    staff = models.ForeignKey(Staff, on_delete=models.CASCADE, related_name="photos")
    url = models.CharField(
        default="https://d3hn0m4rbsz438.cloudfront.net/avatar1.png", max_length=512
    )
    title = models.CharField(max_length=255, blank=True, default="")


class StaffRole(Model):
    """StaffRole."""

    class Meta:
        db_table = "canvas_sdk_data_api_staffrole_001"
        constraints = [
            models.UniqueConstraint(
                fields=["staff", "internal_code"],
                name="staff_internal_code_unique",
            )
        ]

    class RoleDomain(TextChoices):
        CLINICAL = "CLI", "Clinical"
        ADMINISTRATIVE = "ADM", "Administrative"
        HYBRID = "HYB", "Hybrid"

        @staticmethod
        def clinical_domains() -> list["StaffRole.RoleDomain"]:
            """Return a list of clinical role domains."""
            return [StaffRole.RoleDomain.CLINICAL, StaffRole.RoleDomain.HYBRID]

    class RoleType(TextChoices):
        NON_LICENSED = "NON-LICENSED", "Non-Licensed"
        LICENSED = "LICENSED", "Licensed"
        PROVIDER = "PROVIDER", "Provider"

    staff = models.ForeignKey(Staff, on_delete=models.CASCADE, related_name="roles")
    internal_code = models.CharField(max_length=10)
    public_abbreviation = models.CharField(max_length=10, default="", blank=True)
    domain = models.CharField(max_length=3, choices=RoleDomain.choices, db_index=True)
    name = models.CharField(max_length=50)
    domain_privilege_level = models.IntegerField(default=0)
    permissions = models.JSONField(default=dict, blank=True, null=True)
    role_type = models.CharField(max_length=50, choices=RoleType.choices, blank=True)


class StaffLicense(IdentifiableModel):
    """StaffLicense."""

    class Meta:
        db_table = "canvas_sdk_data_api_stafflicense_001"

    class LicenseType(TextChoices):
        CLIA = "CLIA", "CLIA"
        DEA = "DEA", "DEA"
        PTAN = "PTAN", "PTAN"
        STATE_LICENSE = "STATE", "State license"
        TAXONOMY = "TAXONOMY", "Taxonomy"
        SPI = "SPI", "SPI"
        OTHER = "OTHER", "Other"

    staff = models.ForeignKey(Staff, on_delete=models.CASCADE, related_name="licenses")
    issuing_authority_long_name = models.CharField(max_length=200, null=True, blank=True)
    issuing_authority_url = models.URLField(blank=True, null=True)
    license_or_certification_identifier = models.CharField(max_length=50)
    issuance_date = models.DateField()
    expiration_date = models.DateField()
    license_type = models.CharField(
        max_length=50, choices=LicenseType.choices, null=True, blank=True
    )
    primary = models.BooleanField(default=False)
    state = models.CharField(max_length=2, blank=True, null=True)


__exports__ = (
    "Staff",
    "StaffAddress",
    "StaffContactPoint",
    "StaffLicense",
    "StaffPhoto",
    "StaffRole",
)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class ReloadPluginsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ReloadPluginsResponse(_message.Message):
    __slots__ = ("success",)
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    def __init__(self, success: bool = ...) -> None: ...

class ReloadPluginRequest(_message.Message):
    __slots__ = ("plugin",)
    PLUGIN_FIELD_NUMBER: _ClassVar[int]
    plugin: str
    def __init__(self, plugin: _Optional[str] = ...) -> None: ...

class ReloadPluginResponse(_message.Message):
    __slots__ = ("success",)
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    def __init__(self, success: bool = ...) -> None: ...

class UnloadPluginRequest(_message.Message):
    __slots__ = ("plugin",)
    PLUGIN_FIELD_NUMBER: _ClassVar[int]
    plugin: str
    def __init__(self, plugin: _Optional[str] = ...) -> None: ...

class UnloadPluginResponse(_message.Message):
    __slots__ = ("success",)
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    def __init__(self, success: bool = ...) -> None: ...



from enum import Enum

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand
from canvas_sdk.commands.constants import ServiceProvider


class ReferCommand(BaseCommand):
    """A class for managing a Refer command within a specific note."""

    class Meta:
        key = "refer"

    class ClinicalQuestion(Enum):
        """Clinical question choices."""

        COGNITIVE_ASSISTANCE = "Cognitive Assistance (Advice/Guidance)"
        ASSISTANCE_WITH_ONGOING_MANAGEMENT = "Assistance with Ongoing Management"
        SPECIALIZED_INTERVENTION = "Specialized intervention"
        DIAGNOSTIC_UNCERTAINTY = "Diagnostic Uncertainty"

    class Priority(Enum):
        """Priority choices."""

        ROUTINE = "Routine"
        URGENT = "Urgent"

    service_provider: ServiceProvider | None = Field(
        default=None, json_schema_extra={"commands_api_name": "refer_to"}
    )
    diagnosis_codes: list[str] = Field(
        default=[], json_schema_extra={"commands_api_name": "indications"}
    )
    clinical_question: ClinicalQuestion | None = None
    priority: Priority | None = None
    notes_to_specialist: str | None = None
    include_visit_note: bool = False
    comment: str | None = None
    linked_items_urns: list[str] | None = None

    @property
    def values(self) -> dict:
        """The Refer command's field values."""
        values = super().values

        if self.service_provider and self.is_dirty("service_provider"):
            values["service_provider"] = self.service_provider.model_dump()

        return values


__exports__ = ("ReferCommand",)
